"""Views route registration for Lenslet."""

from __future__ import annotations

from fastapi import FastAPI

from .server_models import ViewsPayload
from .workspace import Workspace


def register_views_routes(app: FastAPI, workspace: Workspace) -> None:
    @app.get("/views", response_model=ViewsPayload)
    def get_views():
        return workspace.load_views()

    @app.put("/views", response_model=ViewsPayload)
    def put_views(body: ViewsPayload):
        payload = body.model_dump()
        workspace.save_views(payload)
        return body



from __future__ import annotations

import os
import time
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Callable


@dataclass(frozen=True)
class IndexColumns:
    source_values: list[Any]
    path_values: list[Any]
    name_values: list[Any]
    mime_values: list[Any]
    width_values: list[Any]
    height_values: list[Any]
    size_values: list[Any]
    mtime_values: list[Any]
    metrics_values: list[Any]
    metric_columns: tuple[tuple[str, list[Any]], ...]


@dataclass
class ScannedRow:
    row_idx: int
    logical_path: str
    source: str
    folder_norm: str
    item: Any
    discovered_dims: tuple[int, int] | None = None


@dataclass
class ScanResult:
    rows: list[ScannedRow]
    remote_tasks: list[tuple[str, Any, str, str]]
    skipped_local: int


class ProgressTicker:
    def __init__(self, *, total: int, label: str, emit: Callable[[int, int, str], None]) -> None:
        self.total = total
        self.label = label
        self.emit = emit
        self.done = 0
        self.last_print = 0.0

    def step(self) -> None:
        self.done += 1
        now = time.monotonic()
        if now - self.last_print > 0.1 or self.done == self.total:
            self.emit(self.done, self.total, self.label)
            self.last_print = now

    def finish(self) -> None:
        if self.done < self.total:
            self.emit(self.total, self.total, self.label)


def build_table_indexes(
    storage: Any,
    *,
    item_factory: Callable[..., Any],
    index_factory: Callable[..., Any],
) -> None:
    generated_at = datetime.now(timezone.utc).isoformat()
    columns = build_index_columns(storage)
    scan = scan_rows(storage, columns, item_factory=item_factory)
    assemble_indexes(storage, scan.rows, generated_at=generated_at, index_factory=index_factory)

    if scan.remote_tasks:
        storage._probe_remote_dimensions(scan.remote_tasks)
    if scan.skipped_local:
        print(f"[lenslet] Skipped {scan.skipped_local} local path(s) (local sources disabled or invalid).")


def build_index_columns(storage: Any) -> IndexColumns:
    row_count = storage._row_count
    none_values: list[Any] = [None] * row_count

    def values_for(column: str | None) -> list[Any]:
        if not column:
            return none_values
        return storage._data.get(column, none_values)

    return IndexColumns(
        source_values=values_for(storage._source_column),
        path_values=values_for(storage._path_column),
        name_values=values_for(storage._name_column),
        mime_values=values_for(storage._mime_column),
        width_values=values_for(storage._width_column),
        height_values=values_for(storage._height_column),
        size_values=values_for(storage._size_column),
        mtime_values=values_for(storage._mtime_column),
        metrics_values=values_for(storage._metrics_column),
        metric_columns=collect_metric_columns(storage, none_values),
    )


def collect_metric_columns(storage: Any, none_values: list[Any]) -> tuple[tuple[str, list[Any]], ...]:
    used_columns = {
        storage._source_column,
        storage._path_column,
        storage._name_column,
        storage._mime_column,
        storage._width_column,
        storage._height_column,
        storage._size_column,
        storage._mtime_column,
    }

    metric_columns: list[tuple[str, list[Any]]] = []
    for column in storage._columns:
        if column in used_columns:
            continue
        if column.lower() in storage.RESERVED_COLUMNS:
            continue
        metric_columns.append((column, storage._data.get(column, none_values)))
    return tuple(metric_columns)


def scan_rows(storage: Any, columns: IndexColumns, *, item_factory: Callable[..., Any]) -> ScanResult:
    row_count = storage._row_count
    seen_paths: set[str] = set()
    rows: list[ScannedRow] = []
    remote_tasks: list[tuple[str, Any, str, str]] = []
    skipped_local = 0

    progress = ProgressTicker(
        total=row_count,
        label=f"table:{storage._source_column}",
        emit=storage._progress,
    )

    source_values = columns.source_values
    path_values = columns.path_values
    name_values = columns.name_values
    mime_values = columns.mime_values
    width_values = columns.width_values
    height_values = columns.height_values
    size_values = columns.size_values
    mtime_values = columns.mtime_values
    metrics_values = columns.metrics_values

    has_path_column = storage._path_column is not None
    local_prefix = storage._local_prefix
    allow_local = storage._allow_local
    skip_indexing = storage._skip_indexing
    skip_local_realpath_validation = getattr(storage, "_skip_local_realpath_validation", False)

    extract_name = storage._extract_name
    is_supported_image = storage._is_supported_image
    is_s3_uri = storage._is_s3_uri
    is_http_url = storage._is_http_url
    derive_logical_path = storage._derive_logical_path
    normalize_item_path = storage._normalize_item_path
    normalize_path = storage._normalize_path
    dedupe_path = storage._dedupe_path
    guess_mime = storage._guess_mime
    coerce_int = storage._coerce_int
    coerce_timestamp = storage._coerce_timestamp
    coerce_float = storage._coerce_float
    resolve_local_source = storage._resolve_local_source
    resolve_local_source_lexical = getattr(storage, "_resolve_local_source_lexical", None)
    read_dimensions_fast = storage._read_dimensions_fast

    metric_columns = columns.metric_columns
    has_metrics_column = storage._metrics_column is not None

    for idx in range(row_count):
        source_value = source_values[idx]
        if source_value is None:
            progress.step()
            continue
        if isinstance(source_value, os.PathLike):
            source_value = os.fspath(source_value)
        if not isinstance(source_value, str):
            progress.step()
            continue

        source = source_value.strip()
        if not source:
            progress.step()
            continue

        fallback_name = extract_name(source)
        name_value = name_values[idx]
        name = str(name_value).strip() if name_value else fallback_name
        if not is_supported_image(name):
            if is_supported_image(fallback_name):
                name = fallback_name
            else:
                progress.step()
                continue

        logical_value = None
        if has_path_column and not is_s3_uri(source):
            logical_value = path_values[idx]
        logical_path = str(logical_value).strip() if logical_value else derive_logical_path(source)
        if logical_path and os.path.isabs(logical_path) and local_prefix:
            try:
                rel = os.path.relpath(logical_path, local_prefix)
                if not rel.startswith(".."):
                    logical_path = rel
            except ValueError:
                pass

        logical_path = normalize_item_path(logical_path)
        if not logical_path:
            progress.step()
            continue

        logical_path = dedupe_path(logical_path, seen_paths)
        seen_paths.add(logical_path)

        mime_value = mime_values[idx]
        mime = str(mime_value).strip() if mime_value else guess_mime(name or fallback_name or source)

        size = coerce_int(size_values[idx])
        mtime = coerce_timestamp(mtime_values[idx])
        width = coerce_int(width_values[idx])
        height = coerce_int(height_values[idx])

        is_s3 = is_s3_uri(source)
        is_http = is_http_url(source)
        is_local = not is_s3 and not is_http
        resolved_local_source: str | None = None

        if is_local:
            if not allow_local:
                skipped_local += 1
                progress.step()
                continue
            try:
                if skip_local_realpath_validation and callable(resolve_local_source_lexical):
                    resolved_local_source = resolve_local_source_lexical(source)
                else:
                    resolved_local_source = resolve_local_source(source)
            except ValueError:
                skipped_local += 1
                progress.step()
                continue
            if (not skip_local_realpath_validation) and (not os.path.exists(resolved_local_source)):
                print(f"[lenslet] Warning: File not found: {resolved_local_source}")
                progress.step()
                continue
            if size is None:
                try:
                    size = os.path.getsize(resolved_local_source)
                except Exception:
                    size = 0

        if size is None:
            size = 0

        if mtime is None:
            if is_local:
                try:
                    if resolved_local_source is not None:
                        local_source = resolved_local_source
                    elif skip_local_realpath_validation and callable(resolve_local_source_lexical):
                        local_source = resolve_local_source_lexical(source)
                    else:
                        local_source = resolve_local_source(source)
                    mtime = os.path.getmtime(local_source)
                except Exception:
                    mtime = time.time()
            else:
                mtime = time.time()

        discovered_dims = None
        w = width or 0
        h = height or 0
        if (w == 0 or h == 0) and is_local and not skip_indexing:
            try:
                if resolved_local_source is not None:
                    local_source = resolved_local_source
                elif skip_local_realpath_validation and callable(resolve_local_source_lexical):
                    local_source = resolve_local_source_lexical(source)
                else:
                    local_source = resolve_local_source(source)
                dims = read_dimensions_fast(local_source)
                if dims:
                    w, h = dims
                    discovered_dims = dims
            except Exception:
                pass

        metrics: dict[str, float] = {}
        for metric_key, metric_values in metric_columns:
            num = coerce_float(metric_values[idx])
            if num is None:
                continue
            metrics[metric_key] = num

        if has_metrics_column:
            raw_metrics = metrics_values[idx]
            if isinstance(raw_metrics, dict):
                for raw_key, raw_value in raw_metrics.items():
                    num = coerce_float(raw_value)
                    if num is None:
                        continue
                    metrics[str(raw_key)] = num

        item = item_factory(
            path=logical_path,
            name=name,
            mime=mime,
            width=w,
            height=h,
            size=size,
            mtime=mtime or 0.0,
            url=source if is_http else None,
            source=source,
            metrics=metrics,
        )

        folder = os.path.dirname(logical_path).replace("\\", "/")
        folder_norm = normalize_path(folder)

        rows.append(
            ScannedRow(
                row_idx=idx,
                logical_path=logical_path,
                source=source,
                folder_norm=folder_norm,
                item=item,
                discovered_dims=discovered_dims,
            )
        )

        if (is_s3 or is_http) and (w == 0 or h == 0) and not skip_indexing:
            remote_tasks.append((logical_path, item, source, name))

        progress.step()

    progress.finish()
    return ScanResult(rows=rows, remote_tasks=remote_tasks, skipped_local=skipped_local)


def assemble_indexes(
    storage: Any,
    rows: list[ScannedRow],
    *,
    generated_at: str,
    index_factory: Callable[..., Any],
) -> None:
    dir_children: dict[str, set[str]] = {}

    for row in rows:
        storage._items[row.logical_path] = row.item
        storage._source_paths[row.logical_path] = row.source
        storage._row_dimensions[row.row_idx] = (row.item.width, row.item.height)
        storage._path_to_row[row.logical_path] = row.row_idx
        storage._row_to_path[row.row_idx] = row.logical_path

        if row.discovered_dims:
            storage._dimensions[row.logical_path] = row.discovered_dims

        storage._indexes.setdefault(
            row.folder_norm,
            index_factory(
                path="/" + row.folder_norm if row.folder_norm else "/",
                generated_at=generated_at,
                items=[],
                dirs=[],
            ),
        ).items.append(row.item)

        parts = row.folder_norm.split("/") if row.folder_norm else []
        for depth in range(len(parts)):
            parent = "/".join(parts[:depth])
            child = parts[depth]
            dir_children.setdefault(parent, set()).add(child)

    storage._indexes.setdefault(
        "",
        index_factory(path="/", generated_at=generated_at, items=[], dirs=[]),
    )

    for parent, children in dir_children.items():
        index = storage._indexes.setdefault(
            parent,
            index_factory(
                path="/" + parent if parent else "/",
                generated_at=generated_at,
                items=[],
                dirs=[],
            ),
        )
        index.dirs = sorted(children)


def extract_row_metrics(storage: Any, row_idx: int) -> dict[str, float]:
    none_values: list[Any] = [None] * storage._row_count
    metric_columns = collect_metric_columns(storage, none_values)

    metrics: dict[str, float] = {}
    for column, values in metric_columns:
        num = storage._coerce_float(values[row_idx])
        if num is None:
            continue
        metrics[column] = num
    return metrics


def extract_row_metrics_map(storage: Any, row_idx: int) -> dict[str, float]:
    if not storage._metrics_column:
        return {}

    raw = storage._data.get(storage._metrics_column, [None] * storage._row_count)[row_idx]
    if not isinstance(raw, dict):
        return {}

    result: dict[str, float] = {}
    for key, value in raw.items():
        num = storage._coerce_float(value)
        if num is None:
            continue
        result[str(key)] = num
    return result



from __future__ import annotations


def normalize_search_path(path: str) -> str:
    """Normalize a logical path token for search scope/path comparisons."""
    normalized = (path or "").replace("\\", "/").strip()
    if not normalized or normalized == "/":
        return ""
    return normalized.strip("/")


def path_in_scope(*, logical_path: str, scope_norm: str) -> bool:
    """Return True when logical_path is within scope_norm (or equal)."""
    logical_norm = normalize_search_path(logical_path)
    if not scope_norm:
        return True
    return logical_norm == scope_norm or logical_norm.startswith(f"{scope_norm}/")


def build_search_haystack(
    *,
    logical_path: str,
    name: str,
    tags: list[str],
    notes: str,
    source: str | None,
    url: str | None,
    include_source_fields: bool,
) -> str:
    """Build one canonical lowercase search haystack."""
    parts: list[str] = [
        name or "",
        normalize_search_path(logical_path),
        " ".join(str(tag) for tag in tags if tag is not None),
        notes or "",
    ]
    if include_source_fields:
        parts.extend([
            source or "",
            url or "",
        ])
    return " ".join(part for part in parts if part).lower()



from __future__ import annotations

import math
import os
from datetime import datetime
from typing import Any, Callable, Iterable


def resolve_named_column(columns: list[str], candidates: tuple[str, ...]) -> str | None:
    candidates_lower = {candidate.lower() for candidate in candidates}
    for column in columns:
        if column.lower() in candidates_lower:
            return column
    return None


def resolve_column(columns: list[str], name: str) -> str | None:
    for column in columns:
        if column == name:
            return column
    name_lower = name.lower()
    for column in columns:
        if column.lower() == name_lower:
            return column
    return None


def iter_sample(values: list[Any]) -> Iterable[str]:
    for value in values:
        if value is None:
            continue
        if isinstance(value, float) and math.isnan(value):
            continue
        if isinstance(value, os.PathLike):
            value = os.fspath(value)
        if not isinstance(value, str):
            continue
        cleaned = value.strip()
        if not cleaned:
            continue
        yield cleaned


def loadable_score(
    values: list[Any],
    *,
    sample_size: int,
    is_loadable_value: Callable[[str], bool],
) -> tuple[int, int]:
    total = 0
    matches = 0
    for value in iter_sample(values):
        total += 1
        if is_loadable_value(value):
            matches += 1
        if total >= sample_size:
            break
    return total, matches


def resolve_source_column(
    columns: list[str],
    data: dict[str, list[Any]],
    source_column: str | None,
    *,
    loadable_threshold: float,
    sample_size: int,
    allow_local: bool,
    is_loadable_value: Callable[[str], bool],
) -> str:
    if source_column:
        resolved = resolve_column(columns, source_column)
        if resolved is None:
            raise ValueError(f"source column '{source_column}' not found")
        return resolved

    for column in columns:
        total, matches = loadable_score(
            data.get(column, []),
            sample_size=sample_size,
            is_loadable_value=is_loadable_value,
        )
        if total == 0:
            continue
        if matches / total >= loadable_threshold:
            return column

    if not allow_local:
        raise ValueError(
            "No loadable column found. Local file sources are disabled; provide an S3/HTTP column."
        )
    raise ValueError(
        "No loadable column found. Pass source_column explicitly or provide a base_dir for local paths."
    )


def resolve_path_column(
    columns: list[str],
    path_column: str | None,
    *,
    logical_path_columns: tuple[str, ...],
) -> str | None:
    if path_column:
        resolved = resolve_column(columns, path_column)
        if resolved is None:
            raise ValueError(f"path column '{path_column}' not found")
        return resolved
    return resolve_named_column(columns, logical_path_columns)


def coerce_float(value: Any) -> float | None:
    if value is None:
        return None
    if isinstance(value, (int, float)):
        return float(value)
    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def coerce_int(value: Any) -> int | None:
    if value is None:
        return None
    if isinstance(value, int):
        return value
    if isinstance(value, float):
        return int(value)
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def coerce_timestamp(value: Any) -> float | None:
    if value is None:
        return None
    if isinstance(value, (int, float)):
        return float(value)
    if isinstance(value, datetime):
        return float(value.timestamp())
    if hasattr(value, "timestamp"):
        try:
            return float(value.timestamp())
        except Exception:
            return None
    return coerce_float(value)



from __future__ import annotations

import io
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator
from urllib.parse import urlparse
from urllib.request import Request, urlopen

import pyarrow as pa
import pyarrow.parquet as pq


@dataclass(frozen=True)
class EmbedConfig:
    embedding_column: str = "embedding_mobilenet_v3_small"
    model: str = "mobilenet_v3_small"
    batch_size: int = 32
    parquet_batch_size: int = 256
    device: str = "cpu"
    base_dir: str | None = None
    normalize: bool = False
    limit: int | None = None
    error_policy: str = "zero"
    num_workers: int = 8
    show_progress: bool = True


def embed_parquet(
    parquet_path: str | Path,
    image_column: str,
    output_path: str | Path | None = None,
    config: EmbedConfig | None = None,
) -> Path:
    cfg = config or EmbedConfig()
    parquet_path = Path(parquet_path)
    if not parquet_path.exists():
        raise FileNotFoundError(f"Parquet file not found: {parquet_path}")

    if output_path is None:
        stem = parquet_path.with_suffix("").name
        output_path = parquet_path.with_name(f"{stem}_with_{cfg.embedding_column}.parquet")
    output_path = Path(output_path)

    model, preprocess, dim = _build_model(cfg.model, cfg.device)
    embed_type = pa.list_(pa.float32(), dim)

    pf = pq.ParquetFile(str(parquet_path))
    if image_column not in pf.schema.names:
        raise ValueError(f"Column not found: {image_column}")

    total_rows = pf.metadata.num_rows if pf.metadata else None
    if cfg.limit is not None:
        total_rows = min(total_rows or cfg.limit, cfg.limit)

    tqdm = _lazy_import_tqdm() if cfg.show_progress else None
    progress = tqdm(total=total_rows, unit="rows") if tqdm else None

    writer = None
    processed = 0
    for batch in _iter_batches(str(parquet_path), cfg.parquet_batch_size, cfg.limit):
        paths_raw = batch.column(image_column).to_pylist()
        paths = [_resolve_path(p, cfg.base_dir) for p in paths_raw]

        embeddings = _encode_batch(
            paths=paths,
            model=model,
            preprocess=preprocess,
            dim=dim,
            device=cfg.device,
            batch_size=cfg.batch_size,
            error_policy=cfg.error_policy,
            normalize=cfg.normalize,
            num_workers=cfg.num_workers,
        )

        embed_array = pa.array(embeddings.tolist(), type=embed_type)
        out_batch = batch.append_column(cfg.embedding_column, embed_array)
        if writer is None:
            writer = pq.ParquetWriter(str(output_path), out_batch.schema)
        writer.write_batch(out_batch)
        processed += out_batch.num_rows
        if progress:
            progress.update(out_batch.num_rows)

    if writer is not None:
        writer.close()
    if progress:
        progress.close()

    return output_path


def _lazy_import_torch():
    try:
        import torch
        import torchvision
        from torchvision import transforms
        from torchvision.models import (
            MobileNet_V3_Small_Weights,
            ResNet18_Weights,
            EfficientNet_B0_Weights,
            mobilenet_v3_small,
            resnet18,
            efficientnet_b0,
        )
    except Exception as exc:
        raise RuntimeError(
            "Embedding requires torch + torchvision. Install: pip install torch torchvision"
        ) from exc

    return {
        "torch": torch,
        "torchvision": torchvision,
        "transforms": transforms,
        "weights": {
            "mobilenet_v3_small": (MobileNet_V3_Small_Weights, mobilenet_v3_small),
            "resnet18": (ResNet18_Weights, resnet18),
            "efficientnet_b0": (EfficientNet_B0_Weights, efficientnet_b0),
        },
    }


def _lazy_import_pil():
    try:
        from PIL import Image
    except Exception as exc:
        raise RuntimeError("Embedding requires pillow. Install: pip install pillow") from exc
    return Image


def _lazy_import_boto3():
    try:
        import boto3
    except Exception as exc:
        raise RuntimeError("boto3 is required for s3:// URIs. Install: pip install boto3") from exc
    return boto3


def _lazy_import_tqdm():
    try:
        from tqdm import tqdm
    except Exception:
        return None
    return tqdm


def _resolve_path(raw: str, base_dir: str | None) -> str:
    if raw is None:
        return ""
    text = str(raw)
    if text.startswith("s3://") or text.startswith("http://") or text.startswith("https://"):
        return text
    if base_dir and not os.path.isabs(text):
        return str(Path(base_dir) / text)
    return text


def _read_bytes(uri: str) -> bytes:
    if uri.startswith("s3://"):
        boto3 = _lazy_import_boto3()
        parsed = urlparse(uri)
        bucket = parsed.netloc
        key = parsed.path.lstrip("/")
        if not bucket or not key:
            raise ValueError(f"Invalid s3 URI: {uri}")
        s3 = boto3.client("s3")
        obj = s3.get_object(Bucket=bucket, Key=key)
        return obj["Body"].read()
    if uri.startswith("http://") or uri.startswith("https://"):
        req = Request(uri, headers={"User-Agent": "lenslet-embedder/1.0"})
        with urlopen(req, timeout=30) as resp:
            return resp.read()
    with open(uri, "rb") as handle:
        return handle.read()


def _load_image(uri: str):
    Image = _lazy_import_pil()
    raw = _read_bytes(uri)
    image = Image.open(io.BytesIO(raw))
    return image.convert("RGB")


def _build_model(model_name: str, device: str):
    lib = _lazy_import_torch()
    torch = lib["torch"]
    weights_enum, model_fn = lib["weights"][model_name]
    weights = weights_enum.DEFAULT
    model = model_fn(weights=weights)

    if model_name == "mobilenet_v3_small":
        model.classifier = torch.nn.Identity()
    elif model_name == "resnet18":
        model.fc = torch.nn.Identity()
    elif model_name == "efficientnet_b0":
        model.classifier = torch.nn.Identity()

    model.eval().to(device)
    preprocess = weights.transforms()

    input_size = weights.meta.get("input_size", (3, 224, 224))
    dummy = torch.zeros((1, *input_size), device=device)
    with torch.no_grad():
        dim = int(model(dummy).shape[1])

    return model, preprocess, dim


def _iter_batches(parquet_path: str, batch_size: int, limit: int | None) -> Iterator[pa.RecordBatch]:
    pf = pq.ParquetFile(parquet_path)
    remaining = limit
    for batch in pf.iter_batches(batch_size=batch_size):
        if remaining is None:
            yield batch
            continue
        if remaining <= 0:
            break
        if batch.num_rows > remaining:
            yield batch.slice(0, remaining)
            break
        yield batch
        remaining -= batch.num_rows


def _encode_batch(
    paths: list[str],
    model,
    preprocess,
    dim: int,
    device: str,
    batch_size: int,
    error_policy: str,
    normalize: bool,
    num_workers: int,
):
    lib = _lazy_import_torch()
    torch = lib["torch"]
    import numpy as np

    output = np.zeros((len(paths), dim), dtype=np.float32)
    valid_indices: list[int] = []
    valid_tensors = []

    def load_one(uri: str):
        try:
            img = _load_image(uri)
            return preprocess(img), None
        except Exception as exc:
            return None, exc

    if num_workers > 0:
        from concurrent.futures import ThreadPoolExecutor

        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            for idx, (tensor, exc) in enumerate(executor.map(load_one, paths)):
                if exc is not None:
                    if error_policy == "raise":
                        raise RuntimeError(f"Failed to load image: {paths[idx]}") from exc
                    continue
                if tensor is None:
                    continue
                valid_indices.append(idx)
                valid_tensors.append(tensor)
    else:
        for idx, uri in enumerate(paths):
            tensor, exc = load_one(uri)
            if exc is not None:
                if error_policy == "raise":
                    raise RuntimeError(f"Failed to load image: {uri}") from exc
                continue
            if tensor is None:
                continue
            valid_indices.append(idx)
            valid_tensors.append(tensor)

    if valid_tensors:
        for start in range(0, len(valid_tensors), batch_size):
            chunk = valid_tensors[start : start + batch_size]
            idxs = valid_indices[start : start + batch_size]
            batch = torch.stack(chunk).to(device)
            with torch.no_grad():
                vecs = model(batch).detach().cpu().numpy().astype(np.float32)
            output[idxs] = vecs

    if normalize:
        norms = np.linalg.norm(output, axis=1, keepdims=True)
        mask = norms > 0
        output[mask] = output[mask] / norms[mask]

    return output



"""In-memory dataset storage for programmatic API."""
from __future__ import annotations
import hashlib
import os
import struct
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from datetime import datetime, timezone
from io import BytesIO
from threading import Lock
from urllib.parse import urlparse
from PIL import Image
from .progress import ProgressBar
from .search_text import build_search_haystack, normalize_search_path, path_in_scope
from .s3 import S3_DEPENDENCY_ERROR, create_s3_client


@dataclass
class CachedItem:
    """In-memory cached metadata for an image."""
    path: str
    name: str
    mime: str
    width: int
    height: int
    size: int
    mtime: float
    url: str | None = None  # For HTTP/HTTPS sources


@dataclass
class CachedIndex:
    """In-memory cached folder index."""
    path: str
    generated_at: str
    items: list[CachedItem] = field(default_factory=list)
    dirs: list[str] = field(default_factory=list)


class DatasetStorage:
    """
    In-memory storage for programmatic datasets.
    Supports both local file paths and S3 URIs.
    """

    IMAGE_EXTS = (".jpg", ".jpeg", ".png", ".webp")
    REMOTE_HEADER_BYTES = 65536
    REMOTE_DIM_WORKERS = 16

    def __init__(
        self,
        datasets: dict[str, list[str]],
        thumb_size: int = 256,
        thumb_quality: int = 70,
        include_source_in_search: bool = True,
    ):
        """
        Initialize with datasets.
        
        Args:
            datasets: Dict of {dataset_name: [list of paths/URIs]}
            thumb_size: Thumbnail short edge size
            thumb_quality: WebP quality for thumbnails
        """
        self.datasets = datasets
        self.thumb_size = thumb_size
        self.thumb_quality = thumb_quality
        self._include_source_in_search = include_source_in_search
        self._progress_bar = ProgressBar()
        
        # Build flat path structure: /dataset_name/image_name
        self._items: dict[str, CachedItem] = {}  # path -> item
        self._indexes: dict[str, CachedIndex] = {}
        self._thumbnails: dict[str, bytes] = {}
        self._metadata: dict[str, dict] = {}
        self._dimensions: dict[str, tuple[int, int]] = {}
        self._source_paths: dict[str, str] = {}
        self._s3_client_lock = Lock()
        self._s3_session = None
        self._s3_client = None
        self._s3_client_creations = 0
        
        # Build initial index
        self._build_all_indexes()
        self._browse_signature = self._compute_browse_signature()

    def _compute_browse_signature(self) -> str:
        digest = hashlib.sha256()
        for dataset in sorted(self.datasets.keys()):
            digest.update(dataset.encode("utf-8"))
            for source in sorted(self.datasets[dataset]):
                digest.update(source.encode("utf-8"))
        return digest.hexdigest()

    def _is_s3_uri(self, path: str) -> bool:
        """Check if path is an S3 URI."""
        return path.startswith("s3://")

    def _is_http_url(self, path: str) -> bool:
        """Check if path is an HTTP/HTTPS URL."""
        return path.startswith("http://") or path.startswith("https://")

    def _get_s3_client(self):
        with self._s3_client_lock:
            if self._s3_client is not None:
                return self._s3_client
            self._s3_session, self._s3_client = create_s3_client()
            self._s3_client_creations += 1
            return self._s3_client

    def s3_client_creations(self) -> int:
        return self._s3_client_creations

    def _get_presigned_url(self, s3_uri: str, expires_in: int = 3600) -> str:
        """Convert S3 URI to a presigned HTTPS URL using boto3."""
        try:
            from botocore.exceptions import BotoCoreError, ClientError, NoCredentialsError
        except ImportError as exc:  # pragma: no cover - optional dependency
            raise ImportError(S3_DEPENDENCY_ERROR) from exc

        parsed = urlparse(s3_uri)
        bucket = parsed.netloc
        key = parsed.path.lstrip("/")
        if not bucket or not key:
            raise ValueError(f"Invalid S3 URI: {s3_uri}")

        try:
            s3_client = self._get_s3_client()
            return s3_client.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket, "Key": key},
                ExpiresIn=expires_in,
            )
        except (BotoCoreError, ClientError, NoCredentialsError) as e:
            raise RuntimeError(f"Failed to presign S3 URI: {e}") from e

    def _is_supported_image(self, name: str) -> bool:
        """Check if file is a supported image."""
        return name.lower().endswith(self.IMAGE_EXTS)

    def _extract_name(self, path: str) -> str:
        """Extract filename from local path or S3 URI."""
        if self._is_s3_uri(path) or self._is_http_url(path):
            parsed = urlparse(path)
            return os.path.basename(parsed.path)
        return os.path.basename(path)

    def _read_dimensions_from_bytes(self, data: bytes, ext: str | None) -> tuple[int, int] | None:
        """Read image dimensions from in-memory bytes."""
        if not data:
            return None

        kind = None
        if ext in ("jpg", "jpeg"):
            kind = "jpeg"
        elif ext == "png":
            kind = "png"
        elif ext == "webp":
            kind = "webp"
        else:
            if data.startswith(b"\xff\xd8"):
                kind = "jpeg"
            elif data.startswith(b"\x89PNG\r\n\x1a\n"):
                kind = "png"
            elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
                kind = "webp"

        try:
            buf = BytesIO(data)
            if kind == "jpeg":
                return self._jpeg_dimensions(buf)
            if kind == "png":
                return self._png_dimensions(buf)
            if kind == "webp":
                return self._webp_dimensions(buf)
        except Exception:
            return None
        return None

    def _parse_content_range(self, header: str) -> int | None:
        """Parse Content-Range and return total size if present."""
        # Expected format: "bytes start-end/total" or "bytes */total"
        try:
            if "/" not in header:
                return None
            total = header.split("/")[-1].strip()
            if total == "*":
                return None
            return int(total)
        except Exception:
            return None

    def _get_remote_header_bytes(
        self,
        url: str,
        max_bytes: int | None = None,
    ) -> tuple[bytes | None, int | None]:
        """Fetch the first N bytes of a remote image via Range request."""
        max_bytes = max_bytes or self.REMOTE_HEADER_BYTES
        try:
            import urllib.request
            req = urllib.request.Request(
                url,
                headers={"Range": f"bytes=0-{max_bytes - 1}"},
            )
            with urllib.request.urlopen(req) as response:
                data = response.read(max_bytes)
                total = None
                content_range = response.headers.get("Content-Range")
                if content_range:
                    total = self._parse_content_range(content_range)
                if total is None:
                    content_length = response.headers.get("Content-Length")
                    if content_length:
                        try:
                            total = int(content_length)
                        except Exception:
                            total = None
                return data, total
        except Exception:
            return None, None

    def _get_remote_header_info(
        self,
        url: str,
        name: str,
    ) -> tuple[tuple[int, int] | None, int | None]:
        """Try to read dimensions and size from a ranged remote request."""
        header, total = self._get_remote_header_bytes(url)
        if not header:
            return None, total
        ext = os.path.splitext(name)[1].lower().lstrip(".") or None
        return self._read_dimensions_from_bytes(header, ext), total

    def _get_remote_dimensions(self, url: str, name: str) -> tuple[int, int] | None:
        """Try to read dimensions from a ranged remote request."""
        dims, _ = self._get_remote_header_info(url, name)
        return dims

    def _progress(self, done: int, total: int, label: str) -> None:
        self._progress_bar.update(done, total, label)

    def indexing_progress(self) -> dict[str, int | str | bool | None]:
        return self._progress_bar.snapshot()

    def browse_generation(self) -> int:
        return 0

    def browse_cache_signature(self) -> str:
        return self._browse_signature

    def _effective_remote_workers(self, total: int) -> int:
        if total <= 0:
            return 0
        cpu = os.cpu_count() or 1
        return max(1, min(self.REMOTE_DIM_WORKERS, cpu, total))

    def _probe_remote_dimensions(self, tasks: list[tuple[str, CachedItem, str, str]], label: str) -> None:
        """Fetch remote dimensions in parallel and update cached items."""
        total = len(tasks)
        if total == 0:
            return
        workers = self._effective_remote_workers(total)
        done = 0
        last_print = 0.0
        progress_label = label

        def _work(task: tuple[str, CachedItem, str, str]):
            logical_path, item, source_path, name = task
            url = None
            if self._is_s3_uri(source_path):
                try:
                    url = self._get_presigned_url(source_path)
                except Exception:
                    url = None
            elif self._is_http_url(source_path):
                url = source_path
            if not url:
                return logical_path, item, None, None
            dims, total_size = self._get_remote_header_info(url, name)
            return logical_path, item, dims, total_size

        with ThreadPoolExecutor(max_workers=workers) as executor:
            futures = [executor.submit(_work, task) for task in tasks]
            for future in as_completed(futures):
                logical_path, item, dims, total_size = future.result()
                if dims:
                    self._dimensions[logical_path] = dims
                    item.width, item.height = dims
                if total_size:
                    item.size = total_size
                done += 1
                now = time.monotonic()
                if now - last_print > 0.1 or done == total:
                    self._progress(done, total, progress_label)
                    last_print = now

    def _build_all_indexes(self):
        """Build indexes for all datasets."""
        # Root index contains dataset folders
        root_dirs = list(self.datasets.keys())
        self._indexes["/"] = CachedIndex(
            path="/",
            generated_at=datetime.now(timezone.utc).isoformat(),
            items=[],
            dirs=root_dirs,
        )
        
        # Build index for each dataset
        for dataset_name, paths in self.datasets.items():
            items = []
            remote_tasks: list[tuple[str, CachedItem, str, str]] = []
            for source_path in paths:
                name = self._extract_name(source_path)
                if not self._is_supported_image(name):
                    continue
                
                # Create logical path: /dataset_name/filename
                logical_path = f"/{dataset_name}/{name}"
                
                # Determine source type
                is_s3 = self._is_s3_uri(source_path)
                is_http = self._is_http_url(source_path)
                url = None
                size = 0

                if is_s3:
                    # For S3, presign on-demand only
                    size = 0
                elif is_http:
                    # Plain HTTP/HTTPS URL — use as-is
                    url = source_path
                    size = 0
                else:
                    # For local, validate and get size
                    if not os.path.exists(source_path):
                        print(f"[lenslet] Warning: File not found: {source_path}")
                        continue
                    try:
                        size = os.path.getsize(source_path)
                    except Exception:
                        size = 0
                
                mime = self._guess_mime(name)
                mtime = time.time()
                width, height = 0, 0

                # Create cached item
                item = CachedItem(
                    path=logical_path,
                    name=name,
                    mime=mime,
                    width=width,
                    height=height,
                    size=size,
                    mtime=mtime,
                    url=url,  # For S3 images
                )

                items.append(item)
                self._items[logical_path] = item
                self._source_paths[logical_path] = source_path
                if url:
                    remote_tasks.append((logical_path, item, url, name))
                elif is_s3:
                    remote_tasks.append((logical_path, item, source_path, name))
            
            if remote_tasks:
                self._probe_remote_dimensions(remote_tasks, f"remote headers: {dataset_name}")

            # Create dataset index
            dataset_path = f"/{dataset_name}"
            self._indexes[dataset_path] = CachedIndex(
                path=dataset_path,
                generated_at=datetime.now(timezone.utc).isoformat(),
                items=items,
                dirs=[],
            )

    def _normalize_path(self, path: str) -> str:
        """Normalize path for consistent cache keys."""
        p = path.strip("/")
        return "/" + p if p else "/"

    def _canonical_meta_key(self, path: str) -> str:
        """Canonical key for metadata maps (leading slash, no trailing)."""
        p = (path or "").replace("\\", "/").strip()
        if not p:
            return "/"
        p = "/" + p.lstrip("/")
        if p != "/":
            p = p.rstrip("/")
        return p

    def get_index(self, path: str) -> CachedIndex:
        """Get cached index for a folder."""
        norm = self._normalize_path(path)
        if norm in self._indexes:
            return self._indexes[norm]
        raise FileNotFoundError(f"Dataset not found: {path}")

    def validate_image_path(self, path: str) -> None:
        """Ensure path is valid and exists in our dataset."""
        if not path or path not in self._items:
            raise FileNotFoundError(path)

    def get_source_path(self, logical_path: str) -> str:
        """Get original source path/URI for a logical path."""
        if logical_path not in self._source_paths:
            raise FileNotFoundError(logical_path)
        return self._source_paths[logical_path]

    def read_bytes(self, path: str) -> bytes:
        """Read file contents. For S3, downloads from presigned URL. For local, reads file."""
        source_path = self.get_source_path(path)
        item = self._items[path]
        
        if self._is_s3_uri(source_path):  # S3 image
            import urllib.request
            try:
                url = self._get_presigned_url(source_path)
                with urllib.request.urlopen(url) as response:
                    return response.read()
            except Exception as e:
                raise RuntimeError(f"Failed to download from S3: {e}")
        if item.url:  # HTTP/HTTPS image
            import urllib.request
            try:
                with urllib.request.urlopen(item.url) as response:
                    return response.read()
            except Exception as e:
                raise RuntimeError(f"Failed to download from URL: {e}")
        else:  # Local file
            with open(source_path, "rb") as f:
                return f.read()

    def exists(self, path: str) -> bool:
        """Check if path exists in dataset."""
        return path in self._items

    def size(self, path: str) -> int:
        """Get file size."""
        if path in self._items:
            return self._items[path].size
        return 0

    def join(self, *parts: str) -> str:
        """Join path parts."""
        return "/" + "/".join(p.strip("/") for p in parts if p.strip("/"))

    def etag(self, path: str) -> str | None:
        """Get ETag for caching."""
        if path in self._items:
            item = self._items[path]
            return f"{int(item.mtime)}-{item.size}"
        return None

    def get_thumbnail(self, path: str) -> bytes | None:
        """Get thumbnail, generating if needed."""
        if path in self._thumbnails:
            return self._thumbnails[path]

        try:
            raw = self.read_bytes(path)
            thumb, dims = self._make_thumbnail(raw)
            self._thumbnails[path] = thumb
            if dims:
                self._dimensions[path] = dims
            return thumb
        except Exception as e:
            print(f"[lenslet] Failed to generate thumbnail for {path}: {e}")
            return None

    def _make_thumbnail(self, img_bytes: bytes) -> tuple[bytes, tuple[int, int] | None]:
        """Generate a WebP thumbnail. Returns (thumb_bytes, (w, h))."""
        with Image.open(BytesIO(img_bytes)) as im:
            w, h = im.size
            short = min(w, h)
            if short > self.thumb_size:
                scale = self.thumb_size / short
                new_w = max(1, int(w * scale))
                new_h = max(1, int(h * scale))
                im = im.convert("RGB").resize((new_w, new_h), Image.LANCZOS)
            else:
                im = im.convert("RGB")
            out = BytesIO()
            im.save(out, format="WEBP", quality=self.thumb_quality, method=6)
            return out.getvalue(), (w, h)

    def get_dimensions(self, path: str) -> tuple[int, int]:
        """Get image dimensions, loading lazily if needed."""
        if path in self._dimensions:
            return self._dimensions[path]

        item = self._items.get(path)
        if item:
            url = None
            source_path = self.get_source_path(path)
            if self._is_s3_uri(source_path):
                try:
                    url = self._get_presigned_url(source_path)
                except Exception:
                    url = None
            else:
                url = item.url
            if url:
                dims, total = self._get_remote_header_info(url, item.name)
                if total:
                    self._items[path].size = total
                if dims:
                    self._dimensions[path] = dims
                    self._items[path].width = dims[0]
                    self._items[path].height = dims[1]
                    return dims

        try:
            raw = self.read_bytes(path)
            with Image.open(BytesIO(raw)) as im:
                w, h = im.size
                self._dimensions[path] = (w, h)
                
                # Update item dimensions
                if path in self._items:
                    self._items[path].width = w
                    self._items[path].height = h
                
                return w, h
        except Exception:
            return 0, 0

    def get_metadata(self, path: str) -> dict:
        """Get metadata for an image."""
        key = self._canonical_meta_key(path)
        if key in self._metadata:
            return self._metadata[key]
        
        # Get dimensions if available
        w, h = self._dimensions.get(key, (0, 0))
        if (w == 0 or h == 0) and key in self._items:
            w = self._items[key].width
            h = self._items[key].height
        
        meta = {
            "width": w,
            "height": h,
            "tags": [],
            "notes": "",
            "star": None,
            "version": 1,
            "updated_at": "",
            "updated_by": "server",
        }
        self._metadata[key] = meta
        return meta

    def set_metadata(self, path: str, meta: dict) -> None:
        """Update in-memory metadata (session-only)."""
        key = self._canonical_meta_key(path)
        self._metadata[key] = meta

    def search(self, query: str = "", path: str = "/", limit: int = 100) -> list[CachedItem]:
        """Simple in-memory search."""
        q = (query or "").lower()
        scope_norm = normalize_search_path(path)
        
        results = []
        for item in self._items.values():
            if not path_in_scope(logical_path=item.path, scope_norm=scope_norm):
                continue
            meta = self.get_metadata(item.path)
            source = self._source_paths.get(item.path) if self._include_source_in_search else None
            haystack = build_search_haystack(
                logical_path=item.path,
                name=item.name,
                tags=meta.get("tags", []),
                notes=meta.get("notes", ""),
                source=source,
                url=item.url,
                include_source_fields=self._include_source_in_search,
            )
            
            if q in haystack:
                results.append(item)
                if len(results) >= limit:
                    break
        
        return results

    def _jpeg_dimensions(self, f) -> tuple[int, int] | None:
        """Read JPEG dimensions from SOF marker."""
        f.seek(0)
        if f.read(2) != b"\xff\xd8":
            return None
        while True:
            marker = f.read(2)
            if len(marker) < 2 or marker[0] != 0xff:
                return None
            if marker[1] == 0xd9:  # EOI
                return None
            if 0xc0 <= marker[1] <= 0xcf and marker[1] not in (0xc4, 0xc8, 0xcc):
                length_bytes = f.read(2)
                if len(length_bytes) < 2:
                    return None
                _ = struct.unpack(">H", length_bytes)[0]
                f.read(1)  # precision
                size = f.read(4)
                if len(size) < 4:
                    return None
                h, w = struct.unpack(">HH", size)
                return w, h
            length_bytes = f.read(2)
            if len(length_bytes) < 2:
                return None
            length = struct.unpack(">H", length_bytes)[0]
            if length < 2:
                return None
            f.seek(length - 2, 1)

    def _png_dimensions(self, f) -> tuple[int, int] | None:
        """Read PNG dimensions from IHDR chunk."""
        f.seek(0)
        if f.read(8) != b"\x89PNG\r\n\x1a\n":
            return None
        if len(f.read(4)) < 4:
            return None
        if f.read(4) != b"IHDR":
            return None
        data = f.read(8)
        if len(data) < 8:
            return None
        w, h = struct.unpack(">II", data)
        return w, h

    def _webp_dimensions(self, f) -> tuple[int, int] | None:
        """Read WebP dimensions from header."""
        f.seek(0)
        if f.read(4) != b"RIFF":
            return None
        if len(f.read(4)) < 4:
            return None
        if f.read(4) != b"WEBP":
            return None
        chunk = f.read(4)
        if chunk == b"VP8 ":
            if len(f.read(4)) < 4:
                return None
            f.read(3)
            if f.read(3) != b"\x9d\x01\x2a":
                return None
            data = f.read(4)
            if len(data) < 4:
                return None
            w = (data[0] | (data[1] << 8)) & 0x3FFF
            h = (data[2] | (data[3] << 8)) & 0x3FFF
            return w, h
        if chunk == b"VP8L":
            if len(f.read(4)) < 4:
                return None
            if f.read(1) != b"\x2f":
                return None
            data = f.read(4)
            if len(data) < 4:
                return None
            val = struct.unpack("<I", data)[0]
            w = (val & 0x3FFF) + 1
            h = ((val >> 14) & 0x3FFF) + 1
            return w, h
        if chunk == b"VP8X":
            if len(f.read(4)) < 4:
                return None
            f.read(4)
            data = f.read(6)
            if len(data) < 6:
                return None
            w = (data[0] | (data[1] << 8) | (data[2] << 16)) + 1
            h = (data[3] | (data[4] << 8) | (data[5] << 16)) + 1
            return w, h
        return None

    def _guess_mime(self, name: str) -> str:
        """Guess MIME type from filename."""
        n = name.lower()
        if n.endswith(".webp"):
            return "image/webp"
        if n.endswith(".png"):
            return "image/png"
        return "image/jpeg"



from __future__ import annotations

import os
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Callable


def effective_remote_workers(
    total: int,
    *,
    baseline_workers: int,
    max_workers: int,
    cpu_count: Callable[[], int | None] = os.cpu_count,
) -> int:
    if total <= 0:
        return 0
    cpu = cpu_count() or 1
    cap = max(baseline_workers, cpu)
    cap = min(cap, max_workers)
    return max(1, min(cap, total))


def parse_content_range(header: str) -> int | None:
    try:
        if "/" not in header:
            return None
        total = header.split("/")[-1].strip()
        if total == "*":
            return None
        return int(total)
    except Exception:
        return None


def get_remote_header_bytes(
    url: str,
    *,
    max_bytes: int,
    parse_content_range_fn: Callable[[str], int | None] = parse_content_range,
) -> tuple[bytes | None, int | None]:
    try:
        import urllib.request

        req = urllib.request.Request(
            url,
            headers={"Range": f"bytes=0-{max_bytes - 1}"},
        )
        with urllib.request.urlopen(req) as response:
            data = response.read(max_bytes)
            total = None
            content_range = response.headers.get("Content-Range")
            if content_range:
                total = parse_content_range_fn(content_range)
            if total is None:
                content_length = response.headers.get("Content-Length")
                if content_length:
                    try:
                        total = int(content_length)
                    except Exception:
                        total = None
            return data, total
    except Exception:
        return None, None


def get_remote_header_info(
    url: str,
    name: str,
    *,
    max_bytes: int,
    read_dimensions_from_bytes: Callable[[bytes, str | None], tuple[int, int] | None],
    get_remote_header_bytes_fn: Callable[[str, int | None], tuple[bytes | None, int | None]] | None = None,
) -> tuple[tuple[int, int] | None, int | None]:
    if get_remote_header_bytes_fn is None:
        header, total = get_remote_header_bytes(url, max_bytes=max_bytes)
    else:
        header, total = get_remote_header_bytes_fn(url, max_bytes)
    if not header:
        return None, total
    ext = os.path.splitext(name)[1].lower().lstrip(".") or None
    return read_dimensions_from_bytes(header, ext), total


def probe_remote_dimensions(storage: Any, tasks: list[tuple[str, Any, str, str]]) -> None:
    total = len(tasks)
    if total == 0:
        return

    workers = storage._effective_remote_workers(total)
    if workers <= 0:
        return

    def _work(task: tuple[str, Any, str, str]):
        logical_path, item, source_path, name = task
        url = source_path
        if storage._is_s3_uri(source_path):
            try:
                url = storage._get_presigned_url(source_path)
            except Exception:
                url = None
        if not url:
            return logical_path, item, None, None
        dims, total_size = storage._get_remote_header_info(url, name)
        return logical_path, item, dims, total_size

    done = 0
    last_print = 0.0
    progress_label = "remote headers"
    with ThreadPoolExecutor(max_workers=workers) as executor:
        futures = [executor.submit(_work, task) for task in tasks]
        for future in as_completed(futures):
            logical_path, item, dims, total_size = future.result()
            if dims:
                storage._dimensions[logical_path] = dims
                item.width, item.height = dims
            if total_size:
                item.size = total_size
            row_idx = storage._path_to_row.get(logical_path)
            if row_idx is not None:
                storage._row_dimensions[row_idx] = (item.width, item.height)
            done += 1
            now = time.monotonic()
            if now - last_print > 0.1 or done == total:
                storage._progress(done, total, progress_label)
                last_print = now



"""Shared indexing lifecycle contracts for health and CLI consumers."""

from __future__ import annotations

from datetime import datetime, timezone
from threading import Lock
from typing import Any, Callable, Literal

IndexingState = Literal["idle", "running", "ready", "error"]
IndexingListener = Callable[[dict[str, Any]], None]
_TERMINAL_STATES = frozenset({"ready", "error"})
_POLLING_STATES = frozenset({"idle", "running"})


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def coerce_progress_count(value: Any) -> int | None:
    if isinstance(value, bool):
        return None
    if isinstance(value, int) and value >= 0:
        return value
    return None


def normalize_indexing_payload(payload: Any) -> dict[str, Any] | None:
    if not isinstance(payload, dict):
        return None
    state = payload.get("state")
    if state not in {"idle", "running", "ready", "error"}:
        return None

    done = coerce_progress_count(payload.get("done"))
    total = coerce_progress_count(payload.get("total"))
    if done is not None and total is not None and done > total:
        done = total

    normalized: dict[str, Any] = {"state": state}
    scope = payload.get("scope")
    if isinstance(scope, str):
        normalized["scope"] = scope
    if done is not None:
        normalized["done"] = done
    if total is not None:
        normalized["total"] = total
    started_at = payload.get("started_at")
    if isinstance(started_at, str):
        normalized["started_at"] = started_at
    finished_at = payload.get("finished_at")
    if isinstance(finished_at, str):
        normalized["finished_at"] = finished_at
    error = payload.get("error")
    if isinstance(error, str):
        normalized["error"] = error
    return normalized


def indexing_state_is_terminal(state: str) -> bool:
    return state in _TERMINAL_STATES


def indexing_state_requires_poll(state: str) -> bool:
    return state in _POLLING_STATES


def _scope_suffix(payload: dict[str, Any]) -> str:
    scope = payload.get("scope")
    if not isinstance(scope, str) or scope == "/":
        return ""
    return f" ({scope})"


def _progress_text(payload: dict[str, Any]) -> str:
    done = payload.get("done")
    total = payload.get("total")
    if isinstance(done, int) and isinstance(total, int):
        return f"{done}/{total}"
    return ""


def format_cli_indexing_message(payload: Any) -> str | None:
    normalized = normalize_indexing_payload(payload)
    if normalized is None:
        return None

    state = normalized["state"]
    scope = _scope_suffix(normalized)
    progress = _progress_text(normalized)
    if state == "running":
        if progress:
            return f"Startup indexing in progress{scope}: {progress}."
        return f"Startup indexing in progress{scope}."
    if state == "ready":
        if progress:
            return f"Startup indexing ready{scope}: {progress}."
        return f"Startup indexing ready{scope}."
    if state == "error":
        detail = normalized.get("error")
        if isinstance(detail, str) and detail:
            return f"Startup indexing failed{scope}: {detail}"
        return f"Startup indexing failed{scope}."
    return f"Startup indexing idle{scope}."


class CliIndexingReporter:
    """Emit deterministic CLI lifecycle lines from normalized indexing payloads."""

    def __init__(self, *, write: Callable[[str], None] = print) -> None:
        self._write = write
        self._last_state: IndexingState | None = None
        self._last_error: str | None = None
        self._saw_running = False

    def handle_update(self, payload: Any) -> None:
        normalized = normalize_indexing_payload(payload)
        if normalized is None:
            return

        state = normalized["state"]
        if state == "running":
            self._saw_running = True
            if self._last_state == "running":
                return
            self._emit(normalized, state)
            return

        if state == "ready":
            if not self._saw_running or self._last_state == "ready":
                self._last_state = "ready"
                self._last_error = None
                return
            self._emit(normalized, state)
            return

        if state == "error":
            detail = normalized.get("error") if isinstance(normalized.get("error"), str) else None
            if self._last_state == "error" and detail == self._last_error:
                return
            self._emit(normalized, state)
            self._last_error = detail
            return

        self._last_state = "idle"
        self._last_error = None

    def _emit(self, payload: dict[str, Any], state: IndexingState) -> None:
        message = format_cli_indexing_message(payload)
        self._last_state = state
        self._last_error = None
        if message:
            self._write(f"[lenslet] {message}")


class IndexingLifecycle:
    """Thread-safe tracker for startup indexing lifecycle state."""

    def __init__(self, *, scope: str = "/", initial_state: IndexingState = "idle") -> None:
        self._scope = scope
        self._state: IndexingState = initial_state
        self._started_at: str | None = None
        self._finished_at: str | None = None
        self._error: str | None = None
        self._listeners: list[IndexingListener] = []
        self._lock = Lock()

        if initial_state == "ready":
            now = _now_iso()
            self._started_at = now
            self._finished_at = now
        elif initial_state == "error":
            now = _now_iso()
            self._started_at = now
            self._finished_at = now
            self._error = "indexing failed"

    @classmethod
    def ready(cls, *, scope: str = "/") -> "IndexingLifecycle":
        return cls(scope=scope, initial_state="ready")

    def subscribe(self, listener: IndexingListener, *, emit_current: bool = False) -> None:
        payload: dict[str, Any] | None = None
        with self._lock:
            self._listeners.append(listener)
            if emit_current:
                payload = self._snapshot_unlocked()
        if payload is not None:
            self._notify(listener, payload)

    def start(self, *, scope: str | None = None) -> None:
        payload: dict[str, Any]
        with self._lock:
            if scope is not None:
                self._scope = scope
            self._state = "running"
            self._started_at = _now_iso()
            self._finished_at = None
            self._error = None
            payload = self._snapshot_unlocked()
        self._emit(payload)

    def mark_ready(self) -> None:
        payload: dict[str, Any]
        with self._lock:
            now = _now_iso()
            if self._started_at is None:
                self._started_at = now
            self._state = "ready"
            self._finished_at = now
            self._error = None
            payload = self._snapshot_unlocked()
        self._emit(payload)

    def mark_error(self, message: str) -> None:
        payload: dict[str, Any]
        with self._lock:
            now = _now_iso()
            if self._started_at is None:
                self._started_at = now
            self._state = "error"
            self._finished_at = now
            self._error = message
            payload = self._snapshot_unlocked()
        self._emit(payload)

    def snapshot(self, *, done: int | None = None, total: int | None = None) -> dict[str, Any]:
        with self._lock:
            return self._snapshot_unlocked(done=done, total=total)

    def _snapshot_unlocked(self, *, done: int | None = None, total: int | None = None) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "state": self._state,
            "scope": self._scope,
        }
        if done is not None:
            payload["done"] = done
        if total is not None:
            payload["total"] = total
        if self._started_at is not None:
            payload["started_at"] = self._started_at
        if self._finished_at is not None:
            payload["finished_at"] = self._finished_at
        if self._error is not None:
            payload["error"] = self._error
        return payload

    def _emit(self, payload: dict[str, Any]) -> None:
        with self._lock:
            listeners = list(self._listeners)
        for listener in listeners:
            self._notify(listener, payload)

    @staticmethod
    def _notify(listener: IndexingListener, payload: dict[str, Any]) -> None:
        try:
            listener(dict(payload))
        except Exception:
            return



"""Shared S3 helpers for storage backends."""
from __future__ import annotations

from typing import Any


S3_DEPENDENCY_ERROR = (
    "boto3 package required for S3 support. Install with: pip install lenslet[s3]"
)


def create_s3_client() -> tuple[Any | None, Any]:
    """Create an S3 client and return `(session, client)`."""
    try:
        import boto3
    except ImportError as exc:  # pragma: no cover - optional dependency
        raise ImportError(S3_DEPENDENCY_ERROR) from exc

    session_factory = getattr(getattr(boto3, "session", None), "Session", None)
    if callable(session_factory):
        session = session_factory()
        return session, session.client("s3")
    return None, boto3.client("s3")



"""Browse/search/item helpers shared by Lenslet server modules."""

from __future__ import annotations

import io
import threading
import time
from collections import deque
from datetime import datetime, timezone
from typing import Any, Callable

from fastapi import FastAPI, HTTPException, Request

from .browse_cache import RecursiveBrowseCache, RecursiveCachedItemSnapshot
from .metadata import read_jpeg_info, read_png_info, read_webp_info
from .server_models import DirEntry, FolderIndex, ImageMetadataResponse, Item, SearchResult, Sidecar
from .server_sync import _canonical_path, _sidecar_from_meta
from .workspace import Workspace


def _storage_from_request(request: Request):
    return request.state.storage  # type: ignore[attr-defined]


def _ensure_image(storage, path: str) -> None:
    try:
        storage.validate_image_path(path)
    except FileNotFoundError:
        raise HTTPException(404, "file not found")
    except ValueError as exc:
        raise HTTPException(400, str(exc))


def _build_sidecar(storage, path: str) -> Sidecar:
    meta = storage.get_metadata(path)
    return _sidecar_from_meta(meta)


def _build_image_metadata(storage, path: str) -> ImageMetadataResponse:
    mime = storage._guess_mime(path)  # type: ignore[attr-defined]
    if mime not in ("image/png", "image/jpeg", "image/webp"):
        raise HTTPException(415, "metadata reading supports PNG, JPEG, and WebP images only")

    try:
        raw = storage.read_bytes(path)
        if mime == "image/png":
            meta = read_png_info(io.BytesIO(raw))
            fmt = "png"
        elif mime == "image/jpeg":
            meta = read_jpeg_info(io.BytesIO(raw))
            fmt = "jpeg"
        else:
            meta = read_webp_info(io.BytesIO(raw))
            fmt = "webp"
    except HTTPException:
        raise
    except Exception as exc:  # pragma: no cover - unexpected parse errors
        raise HTTPException(500, f"failed to parse metadata: {exc}")

    return ImageMetadataResponse(path=path, format=fmt, meta=meta)


def _build_item(cached, meta: dict, source: str | None = None) -> Item:
    if source is None:
        source = getattr(cached, "source", None)
    canonical = _canonical_path(cached.path)
    metrics = getattr(cached, "metrics", None)
    if metrics is None:
        metrics = meta.get("metrics")
    return Item(
        path=canonical,
        name=cached.name,
        type=cached.mime,
        w=cached.width,
        h=cached.height,
        size=cached.size,
        hasThumb=True,
        hasMeta=True,
        addedAt=datetime.fromtimestamp(cached.mtime, tz=timezone.utc).isoformat(),
        star=meta.get("star"),
        comments=meta.get("notes", ""),
        url=getattr(cached, "url", None),
        source=source,
        metrics=metrics,
    )


def _child_folder_path(parent: str, child: str) -> str:
    parent_path = _canonical_path(parent)
    return _canonical_path(f"{parent_path.rstrip('/')}/{child}")


RECURSIVE_SORT_MODE_SCAN = "scan"
RECURSIVE_CACHE_BUILD_MAX_RETRIES = 2


class HotpathTelemetry:
    """Lightweight in-process counters/timers for hot-path visibility."""

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._counters: dict[str, int] = {}
        self._timers: dict[str, tuple[int, float]] = {}

    def increment(self, key: str, amount: int = 1) -> None:
        if amount == 0:
            return
        with self._lock:
            self._counters[key] = self._counters.get(key, 0) + amount

    def observe_ms(self, key: str, duration_ms: float) -> None:
        if duration_ms < 0:
            duration_ms = 0
        with self._lock:
            count, total = self._timers.get(key, (0, 0.0))
            self._timers[key] = (count + 1, total + duration_ms)

    def snapshot(self, storage=None) -> dict[str, Any]:
        with self._lock:
            counters = dict(self._counters)
            timers = {
                key: {
                    "count": count,
                    "total_ms": round(total_ms, 3),
                    "avg_ms": round(total_ms / count, 3) if count else 0.0,
                }
                for key, (count, total_ms) in self._timers.items()
            }
        s3_creations = _storage_s3_client_creations(storage)
        if s3_creations is not None:
            counters["s3_client_create_total"] = s3_creations
        return {
            "counters": counters,
            "timers_ms": timers,
        }


def _create_hotpath_metrics(app: FastAPI) -> HotpathTelemetry:
    metrics = HotpathTelemetry()
    app.state.hotpath_metrics = metrics
    return metrics


def _labels_health_payload(workspace: Workspace) -> dict[str, Any]:
    if not workspace.can_write:
        return {"enabled": False, "log": None, "snapshot": None}
    return {
        "enabled": True,
        "log": str(workspace.labels_log_path()),
        "snapshot": str(workspace.labels_snapshot_path()),
    }


def _storage_s3_client_creations(storage) -> int | None:
    getter = getattr(storage, "s3_client_creations", None)
    if callable(getter):
        try:
            return int(getter())
        except Exception:
            return None
    raw = getattr(storage, "_s3_client_creations", None)
    if isinstance(raw, int):
        return raw
    return None




def _collect_recursive_cached_items(
    storage,
    root_path: str,
    root_index: Any,
    *,
    cancelled: Callable[[], bool] | None = None,
) -> tuple[list[Any], int]:
    queue: deque[tuple[str, Any]] = deque([(_canonical_path(root_path), root_index)])
    seen_folders: set[str] = set()
    seen_items: set[str] = set()
    items: list[tuple[str, Any]] = []
    total_items = 0

    while queue:
        if cancelled is not None and cancelled():
            break
        folder_path, folder_index = queue.popleft()
        if folder_path in seen_folders:
            continue
        seen_folders.add(folder_path)

        for cached in folder_index.items:
            if cancelled is not None and cancelled():
                break
            item_path = _canonical_path(getattr(cached, "path", ""))
            if item_path in seen_items:
                continue
            seen_items.add(item_path)
            total_items += 1
            items.append((item_path, cached))

        for child_name in folder_index.dirs:
            if cancelled is not None and cancelled():
                break
            child_path = _child_folder_path(folder_path, child_name)
            if child_path in seen_folders:
                continue
            try:
                child_index = storage.get_index(child_path)
            except (FileNotFoundError, ValueError):
                continue
            queue.append((child_path, child_index))

    items.sort(key=lambda pair: pair[0])
    return [cached for _, cached in items], total_items


def _recursive_index_getter(storage) -> Callable[[str], Any]:
    getter = getattr(storage, "get_index_for_recursive", None)
    if callable(getter):
        return getter
    return storage.get_index


def _collect_recursive_count(
    storage,
    root_path: str,
    root_index: Any,
    *,
    cancelled: Callable[[], bool] | None = None,
) -> int:
    queue: deque[tuple[str, Any]] = deque([(_canonical_path(root_path), root_index)])
    seen_folders: set[str] = set()
    seen_items: set[str] = set()
    total_items = 0
    get_index = _recursive_index_getter(storage)

    while queue:
        if cancelled is not None and cancelled():
            break
        folder_path, folder_index = queue.popleft()
        if folder_path in seen_folders:
            continue
        seen_folders.add(folder_path)

        for cached in folder_index.items:
            if cancelled is not None and cancelled():
                break
            item_path = _canonical_path(getattr(cached, "path", ""))
            if item_path in seen_items:
                continue
            seen_items.add(item_path)
            total_items += 1

        for child_name in folder_index.dirs:
            if cancelled is not None and cancelled():
                break
            child_path = _child_folder_path(folder_path, child_name)
            if child_path in seen_folders:
                continue
            try:
                child_index = get_index(child_path)
            except (FileNotFoundError, ValueError):
                continue
            if child_index is None:
                continue
            queue.append((child_path, child_index))

    return total_items


def _recursive_cache_generation_token(storage) -> str:
    parts: list[str] = []
    signature_fn = getattr(storage, "browse_cache_signature", None)
    if callable(signature_fn):
        try:
            signature = str(signature_fn()).strip()
        except Exception:
            signature = ""
        if signature:
            parts.append(signature)

    generation_fn = getattr(storage, "browse_generation", None)
    if callable(generation_fn):
        try:
            generation = str(generation_fn()).strip()
        except Exception:
            generation = ""
        if generation:
            parts.append(generation)

    if not parts:
        return "default"
    return "|".join(parts)


def _load_or_build_recursive_snapshots(
    storage,
    canonical_path: str,
    root_index: Any,
    *,
    sort_mode: str,
    browse_cache: RecursiveBrowseCache,
    defer_persist: bool = False,
    hotpath_metrics: HotpathTelemetry | None = None,
) -> tuple[tuple[RecursiveCachedItemSnapshot, ...], int]:
    generation_token = _recursive_cache_generation_token(storage)
    for _attempt in range(RECURSIVE_CACHE_BUILD_MAX_RETRIES):
        cached_window = browse_cache.load(canonical_path, sort_mode, generation_token)
        if cached_window is not None:
            window, source = cached_window
            if hotpath_metrics is not None:
                hotpath_metrics.increment("folders_recursive_cache_hit_total")
                if source == "disk":
                    hotpath_metrics.increment("folders_recursive_cache_hit_disk_total")
                else:
                    hotpath_metrics.increment("folders_recursive_cache_hit_memory_total")
            return window.items, window.total_items

        if hotpath_metrics is not None:
            hotpath_metrics.increment("folders_recursive_cache_miss_total")

        cached_items, total_items = _collect_recursive_cached_items(
            storage,
            canonical_path,
            root_index,
        )
        snapshots = tuple(
            RecursiveCachedItemSnapshot.from_cached_item(cached)
            for cached in cached_items
        )

        latest_generation = _recursive_cache_generation_token(storage)
        if latest_generation != generation_token:
            if hotpath_metrics is not None:
                hotpath_metrics.increment("folders_recursive_cache_stale_generation_total")
            generation_token = latest_generation
            try:
                root_index = storage.get_index(canonical_path)
            except (FileNotFoundError, ValueError):
                break
            continue

        _window, wrote_persisted = browse_cache.save(
            canonical_path,
            sort_mode,
            generation_token,
            snapshots,
            defer_persist=defer_persist,
        )
        if hotpath_metrics is not None and wrote_persisted:
            hotpath_metrics.increment("folders_recursive_cache_persist_write_total")
        return snapshots, total_items

    cached_items, total_items = _collect_recursive_cached_items(
        storage,
        canonical_path,
        root_index,
    )
    snapshots = tuple(
        RecursiveCachedItemSnapshot.from_cached_item(cached)
        for cached in cached_items
    )
    return snapshots, total_items


def warm_recursive_cache(
    storage,
    path: str,
    browse_cache: RecursiveBrowseCache | None,
    *,
    hotpath_metrics: HotpathTelemetry | None = None,
) -> int:
    if browse_cache is None:
        return 0
    scope_count = getattr(storage, "count_in_scope", None)
    if callable(scope_count):
        return int(scope_count(path))
    try:
        root_index = storage.get_index(path)
    except (FileNotFoundError, ValueError):
        return 0
    canonical_path = _canonical_path(path)
    _snapshots, total_items = _load_or_build_recursive_snapshots(
        storage,
        canonical_path,
        root_index,
        sort_mode=RECURSIVE_SORT_MODE_SCAN,
        browse_cache=browse_cache,
        defer_persist=False,
        hotpath_metrics=hotpath_metrics,
    )
    return total_items


def _build_folder_index(
    storage,
    path: str,
    to_item,
    recursive: bool = False,
    count_only: bool = False,
    browse_cache: RecursiveBrowseCache | None = None,
    hotpath_metrics: HotpathTelemetry | None = None,
) -> FolderIndex:
    try:
        if count_only:
            getter = _recursive_index_getter(storage)
            index = getter(path)
            if index is None:
                raise FileNotFoundError(path)
        else:
            index = storage.get_index(path)
    except ValueError:
        raise HTTPException(400, "invalid path")
    except FileNotFoundError:
        raise HTTPException(404, "folder not found")

    if recursive:
        if hotpath_metrics is not None:
            hotpath_metrics.increment("folders_recursive_requests_total")
        traversal_started = time.perf_counter()
        canonical_path = _canonical_path(path)
        scope_items = getattr(storage, "items_in_scope", None)
        scope_count = getattr(storage, "count_in_scope", None)
        if callable(scope_items):
            if count_only and callable(scope_count):
                total_items = int(scope_count(canonical_path))
                items = []
            else:
                cached_items = scope_items(canonical_path)
                total_items = len(cached_items)
                items = [] if count_only else [to_item(storage, it) for it in cached_items]
        elif count_only:
            total_items = _collect_recursive_count(
                storage,
                canonical_path,
                index,
            )
            items = []
        elif browse_cache is not None:
            snapshots, total_items = _load_or_build_recursive_snapshots(
                storage,
                canonical_path,
                index,
                sort_mode=RECURSIVE_SORT_MODE_SCAN,
                browse_cache=browse_cache,
                defer_persist=True,
                hotpath_metrics=hotpath_metrics,
            )
            items = [to_item(storage, snapshot) for snapshot in snapshots]
        else:
            cached_items, total_items = _collect_recursive_cached_items(
                storage,
                canonical_path,
                index,
            )
            items = [to_item(storage, it) for it in cached_items]
        if hotpath_metrics is not None:
            hotpath_metrics.observe_ms(
                "folders_recursive_traversal_ms",
                (time.perf_counter() - traversal_started) * 1000.0,
            )
            hotpath_metrics.increment("folders_recursive_items_total", total_items)
    else:
        if count_only:
            total_items = len(getattr(index, "items", []) or [])
            items = []
        else:
            items = [to_item(storage, it) for it in index.items]
            total_items = None

    dirs = [DirEntry(name=d, kind="branch") for d in sorted(index.dirs)]

    return FolderIndex(
        path=_canonical_path(path),
        generatedAt=index.generated_at,
        items=items,
        dirs=dirs,
        page=None,
        pageSize=None,
        pageCount=None,
        totalItems=total_items if count_only else None,
    )


def _search_results(storage, to_item, q: str, path: str, limit: int) -> SearchResult:
    hits = storage.search(query=q, path=path, limit=limit)
    return SearchResult(items=[to_item(storage, it) for it in hits])



"""Common API route registration for Lenslet."""

from __future__ import annotations

import asyncio
import threading
from typing import Any, Callable

from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.responses import JSONResponse, StreamingResponse
from PIL import Image
from pydantic import ValidationError

from .server_models import (
    EXPORT_COMPARISON_REQUEST_ADAPTER,
    FolderIndex,
    ImageMetadataResponse,
    SearchResult,
    Sidecar,
    SidecarPatch,
)
from .server_sync import PresenceTracker


RecordUpdateFn = Callable[[str, dict, str], None]


def _update_item(
    storage,
    path: str,
    body: Sidecar,
    updated_by: str,
    *,
    ensure_meta_fields: Callable[[dict], dict],
    now_iso: Callable[[], str],
    sidecar_from_meta: Callable[[dict], Sidecar],
) -> Sidecar:
    meta = storage.get_metadata(path)
    meta = ensure_meta_fields(meta)
    meta["tags"] = body.tags
    meta["notes"] = body.notes
    meta["star"] = body.star
    meta["version"] = meta.get("version", 1) + 1
    meta["updated_at"] = now_iso()
    meta["updated_by"] = updated_by
    storage.set_metadata(path, meta)
    return sidecar_from_meta(meta)


def register_folder_route(
    app: FastAPI,
    to_item,
    *,
    hotpath_metrics: Any | None = None,
) -> None:
    from . import server as _server

    @app.get("/folders", response_model=FolderIndex)
    def get_folder(
        request: Request,
        path: str = "/",
        recursive: bool = False,
        count_only: bool = False,
    ):
        unsupported = [
            name
            for name in ("page", "page_size", "legacy_recursive")
            if name in request.query_params
        ]
        if unsupported:
            unsupported_list = ", ".join(unsupported)
            supported_list = ", ".join(("path", "recursive", "count_only"))
            return _server._error_response(
                400,
                "unsupported_query_params",
                f"unsupported query parameters: {unsupported_list}; "
                f"supported parameters: {supported_list}",
            )
        storage = _server._storage_from_request(request)
        browse_cache = getattr(app.state, "recursive_browse_cache", None)
        return _server._build_folder_index(
            storage,
            _server._canonical_path(path),
            to_item,
            recursive=recursive,
            count_only=count_only,
            browse_cache=browse_cache,
            hotpath_metrics=hotpath_metrics,
        )


def register_common_api_routes(
    app: FastAPI,
    to_item,
    *,
    meta_lock: threading.Lock,
    presence: PresenceTracker,
    broker,
    presence_metrics,
    idempotency_cache,
    record_update: RecordUpdateFn,
    thumb_queue,
    thumb_cache,
    hotpath_metrics,
) -> None:
    from . import server as _server

    register_folder_route(app, to_item, hotpath_metrics=hotpath_metrics)

    def _resolve_image_request(path: str, request: Request):
        storage = _server._storage_from_request(request)
        canonical_path = _server._canonical_path(path)
        _server._ensure_image(storage, canonical_path)
        return storage, canonical_path

    @app.get("/item")
    def get_item(path: str, request: Request):
        storage, path = _resolve_image_request(path, request)
        return _server._build_sidecar(storage, path)

    @app.get("/metadata", response_model=ImageMetadataResponse)
    def get_metadata(path: str, request: Request):
        storage, path = _resolve_image_request(path, request)
        return _server._build_image_metadata(storage, path)

    @app.post("/export-comparison")
    async def export_comparison(request: Request):
        storage = _server._storage_from_request(request)
        try:
            payload = await request.json()
        except Exception:
            return _server._error_response(400, "invalid_json", "request body must be valid JSON")

        try:
            body = EXPORT_COMPARISON_REQUEST_ADAPTER.validate_python(payload)
        except ValidationError as exc:
            return _server._error_response(400, "invalid_request", _server._first_validation_error_detail(exc))

        canonical_paths = [_server._canonical_path(path) for path in body.paths]
        for path in canonical_paths:
            try:
                _server._ensure_image(storage, path)
            except HTTPException as exc:
                return _server._path_validation_error_response(exc)

        try:
            normalized_labels = _server._normalize_export_labels(
                body.labels,
                max_labels=len(canonical_paths),
            )
        except ValueError as exc:
            return _server._error_response(400, "invalid_labels", str(exc))

        ordered_paths, ordered_labels = _server._resolve_export_paths_and_labels(
            canonical_paths,
            normalized_labels,
            body.reverse_order,
        )

        images: list[Image.Image] = []
        source_formats: list[str] = []
        try:
            for path in ordered_paths:
                image, source_format = _server._load_export_image(storage, path)
                images.append(image)
                source_formats.append(source_format)
            exported_png = _server._build_export_png(
                images,
                ordered_labels,
                embed_metadata=body.embed_metadata,
                ordered_paths=ordered_paths,
                source_formats=source_formats,
                reversed_order=body.reverse_order,
            )
        except HTTPException as exc:
            return _server._comparison_export_error_response(exc)
        finally:
            for image in images:
                image.close()

        filename = _server._comparison_export_filename(body.reverse_order)
        headers = {"Content-Disposition": f'attachment; filename="{filename}"'}
        return Response(content=exported_png, media_type="image/png", headers=headers)

    @app.put("/item")
    def put_item(path: str, body: Sidecar, request: Request):
        storage, path = _resolve_image_request(path, request)
        updated_by = _server._updated_by_from_request(request)
        with meta_lock:
            sidecar = _update_item(
                storage,
                path,
                body,
                updated_by,
                ensure_meta_fields=_server._ensure_meta_fields,
                now_iso=_server._now_iso,
                sidecar_from_meta=_server._sidecar_from_meta,
            )
            meta_snapshot = dict(storage.get_metadata(path))
        record_update(path, meta_snapshot)
        client_id = _server._client_id_from_request(request)
        if client_id:
            gallery_id = _server._gallery_id_from_path(path)
            _server._touch_presence_edit(presence, broker, gallery_id, client_id)
        return sidecar

    @app.patch("/item")
    def patch_item(path: str, body: SidecarPatch, request: Request):
        storage, path = _resolve_image_request(path, request)
        idem_key = request.headers.get("Idempotency-Key")
        if not idem_key:
            raise HTTPException(400, "Idempotency-Key header required")
        cached = idempotency_cache.get(idem_key)
        if cached:
            status, payload = cached
            return JSONResponse(status_code=status, content=payload)

        if_match = _server._parse_if_match(request.headers.get("If-Match"))
        if request.headers.get("If-Match") and if_match is None:
            payload = {"error": "invalid_if_match", "message": "If-Match must be an integer version"}
            idempotency_cache.set(idem_key, 400, payload)
            return JSONResponse(status_code=400, content=payload)

        expected = body.base_version
        if if_match is not None:
            if expected is not None and expected != if_match:
                payload = {"error": "version_mismatch", "message": "If-Match and base_version disagree"}
                idempotency_cache.set(idem_key, 400, payload)
                return JSONResponse(status_code=400, content=payload)
            expected = if_match

        if expected is None:
            payload = {"error": "missing_base_version", "message": "base_version or If-Match is required"}
            idempotency_cache.set(idem_key, 400, payload)
            return JSONResponse(status_code=400, content=payload)

        updated = False
        with meta_lock:
            meta = storage.get_metadata(path)
            meta = _server._ensure_meta_fields(meta)
            if expected is not None and expected != meta.get("version", 1):
                current = _server._sidecar_from_meta(meta).model_dump()
                payload = {"error": "version_conflict", "current": current}
                idempotency_cache.set(idem_key, 409, payload)
                return JSONResponse(status_code=409, content=payload)

            updated = _server._apply_patch_to_meta(meta, body)
            if updated:
                meta["version"] = meta.get("version", 1) + 1
                meta["updated_at"] = _server._now_iso()
                meta["updated_by"] = _server._updated_by_from_request(request)
                storage.set_metadata(path, meta)
            meta_snapshot = dict(meta)
        if updated:
            record_update(path, meta_snapshot)
            client_id = _server._client_id_from_request(request)
            if client_id:
                gallery_id = _server._gallery_id_from_path(path)
                _server._touch_presence_edit(presence, broker, gallery_id, client_id)
        sidecar = _server._sidecar_from_meta(meta_snapshot).model_dump()
        idempotency_cache.set(idem_key, 200, sidecar)
        return JSONResponse(status_code=200, content=sidecar)

    _server._register_presence_routes(
        app,
        presence,
        broker,
        metrics=presence_metrics,
    )

    @app.get("/events")
    async def events(request: Request):
        broker.ensure_loop()
        queue = broker.register()
        last_event_id = _server._last_event_id_from_request(request)

        async def event_stream():
            try:
                for record in broker.replay(last_event_id):
                    yield _server._format_sse(record)
                while True:
                    if await request.is_disconnected():
                        break
                    try:
                        record = await asyncio.wait_for(queue.get(), timeout=15)
                    except asyncio.TimeoutError:
                        yield ": ping\n\n"
                        continue
                    yield _server._format_sse(record)
            finally:
                broker.unregister(queue)

        response = StreamingResponse(event_stream(), media_type="text/event-stream")
        response.headers["Cache-Control"] = "no-cache"
        response.headers["Connection"] = "keep-alive"
        return response

    @app.get("/thumb")
    async def get_thumb(path: str, request: Request):
        storage, path = _resolve_image_request(path, request)
        return await _server._thumb_response_async(
            storage,
            path,
            request,
            thumb_queue,
            thumb_cache,
            hotpath_metrics=hotpath_metrics,
        )

    @app.get("/file")
    def get_file(path: str, request: Request):
        storage, path = _resolve_image_request(path, request)
        return _server._file_response(storage, path, request=request, hotpath_metrics=hotpath_metrics)

    @app.get("/search", response_model=SearchResult)
    def search(request: Request, q: str = "", path: str = "/", limit: int = 100):
        storage = _server._storage_from_request(request)
        return _server._search_results(storage, to_item, q, _server._canonical_path(path), limit)



from __future__ import annotations

import base64
import binascii

try:
    import numpy as np
except ImportError:  # pragma: no cover - optional dependency until embeddings are used
    np = None  # type: ignore[assignment]
import pyarrow as pa
import pyarrow.parquet as pq


class EmbeddingIOError(ValueError):
    pass


def _require_numpy() -> None:
    if np is None:  # pragma: no cover - optional dependency
        raise EmbeddingIOError("numpy is required for embedding search")


def _fixed_size_list_array(column: pa.Array | pa.ChunkedArray) -> pa.Array:
    array = column
    if isinstance(array, pa.ChunkedArray):
        array = array.combine_chunks()
    if not pa.types.is_fixed_size_list(array.type):
        raise EmbeddingIOError("embedding column must be a fixed-size list")
    return array


def load_embedding_matrix(parquet_path: str, column: str) -> tuple[np.ndarray, str]:
    _require_numpy()
    table = pq.read_table(parquet_path, columns=[column])
    if column not in table.column_names:
        raise EmbeddingIOError(f"embedding column '{column}' not found")
    array = _fixed_size_list_array(table[column])
    if array.null_count:
        raise EmbeddingIOError("embedding column contains null rows")
    values = array.values
    if values.null_count:
        raise EmbeddingIOError("embedding column contains null values")
    try:
        np_values = values.to_numpy(zero_copy_only=False)
    except Exception as exc:
        raise EmbeddingIOError(f"failed to convert embedding column to numpy: {exc}") from exc
    dim = array.type.list_size
    expected = len(array) * dim
    if np_values.size != expected:
        raise EmbeddingIOError("embedding column length mismatch")
    matrix = np_values.reshape((len(array), dim))
    return np.asarray(matrix, dtype=np.float32), str(values.type)


def decode_base64_vector(encoded: str, dimension: int) -> np.ndarray:
    _require_numpy()
    if not encoded:
        raise EmbeddingIOError("query_vector_b64 is required")
    try:
        raw = base64.b64decode(encoded, validate=True)
    except binascii.Error as exc:
        raise EmbeddingIOError("invalid base64 for query_vector_b64") from exc
    expected_bytes = dimension * 4
    if len(raw) != expected_bytes:
        raise EmbeddingIOError(
            f"query_vector_b64 length mismatch (expected {expected_bytes} bytes for dim {dimension})"
        )
    vec = np.frombuffer(raw, dtype="<f4")
    if vec.size != dimension:
        raise EmbeddingIOError("query_vector_b64 dimension mismatch")
    vec = vec.astype(np.float32, copy=False)
    if not np.isfinite(vec).all():
        raise EmbeddingIOError("query vector contains NaN or infinity")
    norm = float(np.linalg.norm(vec))
    if norm == 0.0 or not np.isfinite(norm):
        raise EmbeddingIOError("query vector has zero magnitude")
    return vec


def normalize_vector(vec: np.ndarray) -> np.ndarray:
    _require_numpy()
    norm = float(np.linalg.norm(vec))
    if norm == 0.0 or not np.isfinite(norm):
        raise EmbeddingIOError("vector has zero magnitude")
    return vec / norm



"""Presence route registration and lifecycle helpers for Lenslet."""

from __future__ import annotations

import asyncio
import contextlib
from typing import Any

from fastapi import FastAPI, HTTPException
from fastapi.responses import JSONResponse

from .server_models import PresenceLeavePayload, PresenceMovePayload, PresencePayload
from .server_runtime import PresenceMetrics
from .server_sync import (
    PresenceCount,
    PresenceLeaseError,
    PresenceScopeError,
    PresenceTracker,
    _canonical_path,
)


def presence_runtime_payload(
    *,
    presence: PresenceTracker,
    broker,
    metrics: PresenceMetrics,
    prune_interval_seconds: float,
) -> dict[str, Any]:
    presence_diag = presence.diagnostics()
    broker_diag = broker.diagnostics()
    metric_diag = metrics.snapshot()
    return {
        "view_ttl_seconds": presence.view_ttl_seconds,
        "edit_ttl_seconds": presence.edit_ttl_seconds,
        "prune_interval_seconds": prune_interval_seconds,
        "active_clients": presence_diag["active_clients"],
        "active_scopes": presence_diag["active_scopes"],
        "stale_pruned_total": presence_diag["stale_pruned_total"],
        "invalid_lease_total": metric_diag["invalid_lease_total"],
        "replay_miss_total": broker_diag["replay_miss_total"],
        "replay_buffer_size": broker_diag["buffer_size"],
        "replay_buffer_capacity": broker_diag["buffer_capacity"],
        "replay_oldest_event_id": broker_diag["oldest_event_id"],
        "replay_newest_event_id": broker_diag["newest_event_id"],
        "connected_sse_clients": broker_diag["connected_sse_clients"],
    }


def presence_count_payload(count: PresenceCount) -> dict[str, int | str]:
    return {
        "gallery_id": count.gallery_id,
        "viewing": count.viewing,
        "editing": count.editing,
    }


def presence_payload_for_client(
    count: PresenceCount,
    client_id: str,
    lease_id: str,
) -> dict[str, int | str]:
    payload: dict[str, int | str] = presence_count_payload(count)
    payload["client_id"] = client_id
    payload["lease_id"] = lease_id
    return payload


def presence_count_for_gallery(counts: list[PresenceCount], gallery_id: str) -> PresenceCount:
    for count in counts:
        if count.gallery_id == gallery_id:
            return count
    return PresenceCount(gallery_id=gallery_id, viewing=0, editing=0)


def publish_presence_counts(broker, counts: list[PresenceCount]) -> None:
    for count in counts:
        broker.publish("presence", presence_count_payload(count))


def publish_presence_deltas(
    broker,
    previous: dict[str, PresenceCount],
    current: dict[str, PresenceCount],
) -> None:
    if previous == current:
        return
    gallery_ids = sorted(set(previous) | set(current))
    for gallery_id in gallery_ids:
        before = previous.get(gallery_id)
        after = current.get(gallery_id)
        before_tuple = (before.viewing, before.editing) if before else (0, 0)
        after_tuple = (after.viewing, after.editing) if after else (0, 0)
        if before_tuple == after_tuple:
            continue
        broker.publish(
            "presence",
            {"gallery_id": gallery_id, "viewing": after_tuple[0], "editing": after_tuple[1]},
        )


def install_presence_prune_loop(
    app: FastAPI,
    presence: PresenceTracker,
    broker,
    interval_seconds: float,
) -> None:
    interval = interval_seconds if interval_seconds > 0 else 5.0
    app.state.presence_tracker = presence
    app.state.presence_prune_interval = interval
    app.state.presence_prune_task = None

    async def _presence_prune_loop() -> None:
        previous = presence.snapshot_counts()
        while True:
            await asyncio.sleep(interval)
            current = presence.snapshot_counts()
            publish_presence_deltas(broker, previous, current)
            previous = current

    async def _start_presence_prune_loop() -> None:
        existing = getattr(app.state, "presence_prune_task", None)
        if existing is not None and not existing.done():
            return
        app.state.presence_prune_task = asyncio.create_task(_presence_prune_loop())

    async def _stop_presence_prune_loop() -> None:
        task = getattr(app.state, "presence_prune_task", None)
        if task is None:
            return
        task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await task
        app.state.presence_prune_task = None

    app.add_event_handler("startup", _start_presence_prune_loop)
    app.add_event_handler("shutdown", _stop_presence_prune_loop)


def presence_invalid_lease_payload(gallery_id: str, client_id: str) -> dict[str, str]:
    return {
        "error": "invalid_lease",
        "gallery_id": gallery_id,
        "client_id": client_id,
    }


def presence_scope_mismatch_payload(
    exc: PresenceScopeError,
    requested_gallery_id: str,
    client_id: str,
) -> dict[str, str]:
    return {
        "error": "scope_mismatch",
        "requested_gallery_id": requested_gallery_id,
        "actual_gallery_id": exc.actual_gallery_id,
        "client_id": client_id,
    }


def touch_presence_edit(presence: PresenceTracker, broker, gallery_id: str, client_id: str) -> None:
    _, counts = presence.touch_edit(gallery_id, client_id)
    publish_presence_counts(broker, counts)


def require_presence_client_id(client_id: str) -> str:
    if not client_id:
        raise HTTPException(400, "client_id required")
    return client_id


def _invalid_lease_response(
    metrics: PresenceMetrics,
    *,
    gallery_id: str,
    client_id: str,
) -> JSONResponse:
    metrics.record_invalid_lease()
    payload = presence_invalid_lease_payload(gallery_id=gallery_id, client_id=client_id)
    return JSONResponse(status_code=409, content=payload)


def _scope_mismatch_response(
    exc: PresenceScopeError,
    *,
    requested_gallery_id: str,
    client_id: str,
) -> JSONResponse:
    payload = presence_scope_mismatch_payload(
        exc,
        requested_gallery_id=requested_gallery_id,
        client_id=client_id,
    )
    return JSONResponse(status_code=409, content=payload)


def register_presence_routes(
    app: FastAPI,
    presence: PresenceTracker,
    broker,
    *,
    metrics: PresenceMetrics,
) -> None:
    def _presence_diag() -> dict[str, Any]:
        prune_interval = float(getattr(app.state, "presence_prune_interval", 5.0))
        return presence_runtime_payload(
            presence=presence,
            broker=broker,
            metrics=metrics,
            prune_interval_seconds=prune_interval,
        )

    @app.get("/presence/diagnostics")
    def presence_diagnostics():
        return _presence_diag()

    @app.post("/presence/join")
    def presence_join(body: PresencePayload):
        gallery_id = _canonical_path(body.gallery_id)
        client_id = require_presence_client_id(body.client_id)
        try:
            lease_id, counts = presence.join(gallery_id, client_id, lease_id=body.lease_id)
        except PresenceLeaseError:
            return _invalid_lease_response(metrics, gallery_id=gallery_id, client_id=client_id)
        publish_presence_counts(broker, counts)
        current = presence_count_for_gallery(counts, gallery_id)
        return presence_payload_for_client(current, client_id, lease_id)

    @app.post("/presence/move")
    def presence_move(body: PresenceMovePayload):
        from_gallery_id = _canonical_path(body.from_gallery_id)
        to_gallery_id = _canonical_path(body.to_gallery_id)
        client_id = require_presence_client_id(body.client_id)
        try:
            response_lease_id = body.lease_id
            counts = presence.move(
                from_gallery_id=from_gallery_id,
                to_gallery_id=to_gallery_id,
                client_id=client_id,
                lease_id=body.lease_id,
            )
        except PresenceLeaseError:
            return _invalid_lease_response(metrics, gallery_id=from_gallery_id, client_id=client_id)
        except PresenceScopeError as exc:
            return _scope_mismatch_response(
                exc,
                requested_gallery_id=from_gallery_id,
                client_id=client_id,
            )
        publish_presence_counts(broker, counts)
        from_scope = presence_count_for_gallery(counts, from_gallery_id)
        to_scope = presence_count_for_gallery(counts, to_gallery_id)
        return {
            "client_id": client_id,
            "lease_id": response_lease_id,
            "from_scope": presence_count_payload(from_scope),
            "to_scope": presence_count_payload(to_scope),
        }

    @app.post("/presence/leave")
    def presence_leave(body: PresenceLeavePayload):
        gallery_id = _canonical_path(body.gallery_id)
        client_id = require_presence_client_id(body.client_id)
        try:
            removed, counts = presence.leave(gallery_id=gallery_id, client_id=client_id, lease_id=body.lease_id)
        except PresenceLeaseError:
            return _invalid_lease_response(metrics, gallery_id=gallery_id, client_id=client_id)
        except PresenceScopeError as exc:
            return _scope_mismatch_response(
                exc,
                requested_gallery_id=gallery_id,
                client_id=client_id,
            )
        publish_presence_counts(broker, counts)
        current = presence_count_for_gallery(counts, gallery_id)
        payload = presence_payload_for_client(current, client_id, body.lease_id)
        payload["removed"] = removed
        return payload

    @app.post("/presence")
    def presence_heartbeat(body: PresencePayload):
        gallery_id = _canonical_path(body.gallery_id)
        client_id = require_presence_client_id(body.client_id)
        try:
            lease_id, counts = presence.touch_view(gallery_id, client_id, lease_id=body.lease_id)
        except PresenceLeaseError:
            return _invalid_lease_response(metrics, gallery_id=gallery_id, client_id=client_id)
        publish_presence_counts(broker, counts)
        current = presence_count_for_gallery(counts, gallery_id)
        return presence_payload_for_client(current, client_id, lease_id)



"""CLI entry point for Lenslet."""
from __future__ import annotations
import argparse
import math
import os
import re
import socket
import subprocess
import sys
import threading
import time
from pathlib import Path
from typing import TYPE_CHECKING, Iterable
from urllib.parse import urlparse

if TYPE_CHECKING:
    from .storage.table import TableStorage


_CLOUDFLARED_URL = (
    "https://github.com/cloudflare/cloudflared/releases/latest/download/"
    "cloudflared-linux-amd64"
)


def _is_remote_uri(value: str) -> bool:
    try:
        parsed = urlparse(value)
    except Exception:
        return False
    return parsed.scheme in {"s3", "http", "https", "hf"}


def _looks_like_hf_dataset(value: str) -> bool:
    if "://" in value:
        return False
    if value.startswith(("/", ".", "~")):
        return False
    parts = value.split("/")
    if len(parts) != 2:
        return False
    org, repo = parts
    repo_base = repo.split("@", 1)[0]
    name_pattern = re.compile(r"^[A-Za-z0-9][A-Za-z0-9._-]{0,127}$")
    return bool(name_pattern.match(org)) and bool(name_pattern.match(repo_base))


def _load_remote_table(uri: str):
    try:
        import unibox as ub
    except ImportError as exc:
        raise RuntimeError(
            "unibox is required to load remote tables. Install with: pip install unibox"
        ) from exc
    table = ub.loads(uri)
    if hasattr(table, "to_pandas"):
        try:
            table = table.to_pandas()
        except TypeError:
            table = table.to_pandas()
    return table


def _port_is_available(host: str, port: int) -> bool:
    try:
        infos = socket.getaddrinfo(host, port, type=socket.SOCK_STREAM)
    except socket.gaierror:
        return False
    for family, socktype, proto, _, sockaddr in infos:
        try:
            with socket.socket(family, socktype, proto) as sock:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind(sockaddr)
                return True
        except OSError:
            continue
    return False


def _find_available_port(host: str, start_port: int = 7070, max_tries: int = 50) -> int:
    for offset in range(max_tries):
        port = start_port + offset
        if _port_is_available(host, port):
            return port
    raise RuntimeError(
        f"No available port found from {start_port} to {start_port + max_tries - 1}."
    )


def _ensure_cloudflared_binary() -> Path:
    binary_path = Path.home() / "cfed"
    if binary_path.is_file():
        if os.access(binary_path, os.X_OK):
            return binary_path
        binary_path.chmod(binary_path.stat().st_mode | 0o111)
        return binary_path

    print("[lenslet] Downloading cloudflared...")
    import urllib.request
    with urllib.request.urlopen(_CLOUDFLARED_URL) as response, binary_path.open("wb") as handle:
        while True:
            chunk = response.read(1024 * 1024)
            if not chunk:
                break
            handle.write(chunk)
    binary_path.chmod(0o755)
    return binary_path


def _tunnel_target_host(bind_host: str) -> str:
    if bind_host in {"127.0.0.1", "0.0.0.0", "localhost", "::1"}:
        return "localhost"
    return bind_host


class _ShareTunnel:
    def __init__(
        self,
        thread: threading.Thread,
        stop_event: threading.Event,
        process_ref: list[subprocess.Popen[str] | None],
        lock: threading.Lock,
    ) -> None:
        self._thread = thread
        self._stop_event = stop_event
        self._process_ref = process_ref
        self._lock = lock

    def stop(self) -> None:
        self._stop_event.set()
        with self._lock:
            process = self._process_ref[0]
        if process is not None:
            _stop_process(process)


def _start_share_tunnel(
    port: int,
    bind_host: str,
    verbose: bool,
    max_retries: int = 3,
    reachable_timeout: float = 20.0,
) -> _ShareTunnel:
    ansi_escape = re.compile(r"\x1b\[[0-9;]*[a-zA-Z]")
    trycloudflare_pattern = re.compile(r"https?://[A-Za-z0-9.-]*trycloudflare\.com\S*")

    def _launch() -> subprocess.Popen[str]:
        binary_path = _ensure_cloudflared_binary()
        target_host = _tunnel_target_host(bind_host)
        cmd = [str(binary_path), "tunnel", "--url", f"{target_host}:{port}"]
        return subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            bufsize=1,
        )

    def _read_share_url(process: subprocess.Popen[str]) -> str | None:
        if process.stdout is None:
            raise RuntimeError("Failed to start cloudflared: no stdout pipe available.")
        for raw in process.stdout:
            line = raw.rstrip()
            if verbose:
                print(f"[cloudflared] {line}")
            cleaned = ansi_escape.sub("", line)
            match = trycloudflare_pattern.search(cleaned)
            if match:
                return match.group(0).rstrip(").,")
        return None

    def _drain_output(lines: Iterable[str], stop_event: threading.Event) -> None:
        for raw in lines:
            if stop_event.is_set():
                break
            if verbose:
                print(f"[cloudflared] {raw.rstrip()}")

    def _wait_for_reachable(url: str, stop_event: threading.Event) -> bool:
        import urllib.error
        import urllib.request

        deadline = time.monotonic() + reachable_timeout
        while True:
            if stop_event.is_set():
                return False
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                return False
            timeout = max(0.5, min(3.0, remaining))
            try:
                request = urllib.request.Request(url, method="HEAD")
                with urllib.request.urlopen(request, timeout=timeout):
                    return True
            except urllib.error.HTTPError:
                return True
            except Exception:
                if time.monotonic() >= deadline:
                    return False
                time.sleep(0.5)

    stop_event = threading.Event()
    process_ref: list[subprocess.Popen[str] | None] = [None]
    lock = threading.Lock()

    def _runner() -> None:
        try:
            last_error = "Share tunnel exited before a URL was created."
            for attempt in range(1, max_retries + 1):
                if stop_event.is_set():
                    return
                process = _launch()
                with lock:
                    process_ref[0] = process
                share_url = _read_share_url(process)
                if share_url is None:
                    if process.poll() not in (None, 0):
                        print("[lenslet] Share tunnel exited before a URL was created.", file=sys.stderr)
                    _stop_process(process)
                    continue
                _print_share_url(share_url)
                if process.stdout is None:
                    raise RuntimeError("Failed to start cloudflared: no stdout pipe available.")
                drain_thread = threading.Thread(
                    target=_drain_output,
                    args=(process.stdout, stop_event),
                    daemon=True,
                )
                drain_thread.start()
                if not _wait_for_reachable(share_url, stop_event):
                    print(
                        "[lenslet] Share URL not reachable within 20 seconds.",
                        file=sys.stderr,
                    )
                return
            if not stop_event.is_set():
                print(f"[lenslet] {last_error}", file=sys.stderr)
        except Exception as exc:
            if not stop_event.is_set():
                print(f"[lenslet] Share tunnel failed: {exc}", file=sys.stderr)

    thread = threading.Thread(target=_runner, daemon=True)
    thread.start()
    return _ShareTunnel(thread, stop_event, process_ref, lock)


def _stop_process(process: subprocess.Popen[str]) -> None:
    if process.poll() is not None:
        return
    process.terminate()
    try:
        process.wait(timeout=5)
    except subprocess.TimeoutExpired:
        process.kill()
        process.wait(timeout=5)


def _print_share_url(url: str) -> None:
    print(f"Share URL: {url}", file=sys.stderr, flush=True)


def main():
    parser = argparse.ArgumentParser(
        prog="lenslet",
        description="Lenslet - Lightweight image gallery server",
        epilog="Example: lenslet ~/Pictures --port 7070",
    )
    parser.add_argument(
        "directory",
        type=str,
        nargs="?",  # Make optional for --version/--help
        help="Directory containing images, a Parquet table, or an HF dataset (org/dataset) to serve",
    )
    parser.add_argument(
        "-p", "--port",
        type=int,
        default=None,
        help="Port to listen on (default: 7070; auto-increment if in use)",
    )
    parser.add_argument(
        "-H", "--host",
        type=str,
        default="127.0.0.1",
        help="Host to bind to (default: 127.0.0.1)",
    )
    parser.add_argument(
        "--thumb-size",
        type=int,
        default=256,
        help="Thumbnail short edge size in pixels (default: 256)",
    )
    parser.add_argument(
        "--thumb-quality",
        type=int,
        default=70,
        help="Thumbnail WEBP quality 1-100 (default: 70)",
    )
    parser.add_argument(
        "--source-column",
        type=str,
        default=None,
        help="Column to load images from in table mode (items.parquet or .parquet file)",
    )
    parser.add_argument(
        "--base-dir",
        type=str,
        default=None,
        help="Base directory for resolving relative paths in table mode",
    )
    parser.add_argument(
        "--no-cache-wh",
        action="store_false",
        dest="cache_wh",
        default=True,
        help="Disable caching width/height back into parquet",
    )
    parser.add_argument(
        "--no-skip-indexing",
        action="store_false",
        dest="skip_indexing",
        default=True,
        help="Probe image dimensions during table load",
    )
    parser.add_argument(
        "--no-thumb-cache",
        action="store_false",
        dest="thumb_cache",
        default=True,
        help="Disable thumbnail cache when a workspace is available",
    )
    parser.add_argument(
        "--no-og-preview",
        action="store_false",
        dest="og_preview",
        default=True,
        help="Disable dataset-based social preview image",
    )
    parser.add_argument(
        "--reload",
        action="store_true",
        help="Enable auto-reload for development",
    )
    parser.add_argument(
        "--no-write",
        action="store_true",
        help="Use a temp workspace under /tmp/lenslet (keeps the source dataset read-only)",
    )
    parser.add_argument(
        "--embedding-column",
        action="append",
        default=None,
        help="Embedding column name (repeatable, comma-separated allowed)",
    )
    parser.add_argument(
        "--embedding-metric",
        action="append",
        default=None,
        help="Embedding metric override in NAME:METRIC form (repeatable)",
    )
    parser.add_argument(
        "--embed",
        action="store_true",
        help="Run CPU embedding inference on a parquet file before launch",
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=32,
        help="Embedding inference batch size (used with --embed)",
    )
    parser.add_argument(
        "--parquet-batch-size",
        type=int,
        default=256,
        help="Rows per parquet batch when embedding (used with --embed)",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=8,
        help="Parallel image loading workers (used with --embed)",
    )
    parser.add_argument(
        "--embedding-preload",
        action="store_true",
        help="Preload embedding indexes on startup",
    )
    parser.add_argument(
        "--embedding-cache",
        action="store_true",
        dest="embedding_cache",
        default=True,
        help="Enable embedding cache (default)",
    )
    parser.add_argument(
        "--no-embedding-cache",
        action="store_false",
        dest="embedding_cache",
        help="Disable embedding cache",
    )
    parser.add_argument(
        "--embedding-cache-dir",
        type=str,
        default=None,
        help="Override embedding cache directory",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Show detailed server logs",
    )
    parser.add_argument(
        "--share",
        action="store_true",
        help="Create a public share URL via cloudflared",
    )
    parser.add_argument(
        "-v", "--version",
        action="store_true",
        help="Show version and exit",
    )

    args = parser.parse_args()

    if args.version:
        from . import __version__
        print(f"lenslet {__version__}")
        sys.exit(0)

    # Directory is required unless --version
    if not args.directory:
        parser.print_help()
        sys.exit(1)

    from .embeddings.config import EmbeddingConfig, parse_embedding_columns, parse_embedding_metrics
    try:
        embedding_config = EmbeddingConfig(
            explicit_columns=parse_embedding_columns(args.embedding_column),
            metric_overrides=parse_embedding_metrics(args.embedding_metric),
        )
    except ValueError as exc:
        print(f"Error: {exc}", file=sys.stderr)
        sys.exit(1)

    # Resolve and validate target
    raw_target = args.directory
    candidate = Path(raw_target).expanduser()
    is_local = candidate.exists()
    is_table_file = False
    is_remote_table = False
    remote_kind = None
    remote_uri = None
    if is_local:
        target = candidate.resolve()
        is_table_file = target.is_file() and target.suffix.lower() == ".parquet"
        if target.is_file():
            if not is_table_file:
                print(f"Error: '{args.directory}' is not a .parquet file", file=sys.stderr)
                sys.exit(1)
        elif not target.is_dir():
            print(f"Error: '{args.directory}' is not a valid directory or .parquet file", file=sys.stderr)
            sys.exit(1)
    elif _is_remote_uri(raw_target):
        is_remote_table = True
        remote_kind = "remote"
        remote_uri = raw_target
    elif _looks_like_hf_dataset(raw_target):
        is_remote_table = True
        remote_kind = "hf"
        remote_uri = f"hf://{raw_target}"
    else:
        print(f"Error: '{args.directory}' does not exist", file=sys.stderr)
        sys.exit(1)

    if args.embed and is_remote_table:
        print("Error: --embed requires a local parquet file", file=sys.stderr)
        sys.exit(1)

    embed_output_path = None
    if args.embed and is_table_file:
        try:
            from .embeddings.embedder import EmbedConfig, embed_parquet
            from .storage.table import load_parquet_schema
        except Exception as exc:
            print(f"Error: {exc}", file=sys.stderr)
            sys.exit(1)

        base_dir = args.base_dir or str(candidate.parent)
        image_column = args.source_column
        if image_column is None:
            try:
                image_column = _detect_source_column(str(candidate), base_dir)
            except Exception as exc:
                print(f"Error: failed to detect image column: {exc}", file=sys.stderr)
                sys.exit(1)
        if image_column is None:
            print("Error: --embed requires --source-column (or a detectable image column)", file=sys.stderr)
            sys.exit(1)

        embed_columns = parse_embedding_columns(args.embedding_column)
        embed_column = embed_columns[0] if embed_columns else "embedding_mobilenet_v3_small"
        if embed_columns and len(embed_columns) > 1:
            print("[lenslet] Warning: --embed uses the first --embedding-column value only.")

        try:
            schema = load_parquet_schema(str(candidate))
            if embed_column in schema.names:
                print(f"[lenslet] Embedding column '{embed_column}' already exists; skipping --embed.")
            else:
                config = EmbedConfig(
                    embedding_column=embed_column,
                    batch_size=args.batch_size,
                    parquet_batch_size=args.parquet_batch_size,
                    num_workers=args.num_workers,
                    base_dir=base_dir,
                )
                embed_output_path = embed_parquet(
                    parquet_path=candidate,
                    image_column=image_column,
                    config=config,
                )
                print(f"[lenslet] Wrote embeddings to {embed_output_path}")
        except Exception as exc:
            print(f"Error: embedding failed: {exc}", file=sys.stderr)
            sys.exit(1)

    if embed_output_path is not None:
        candidate = embed_output_path
        is_table_file = True
        target = candidate.resolve()

    port = args.port
    if port is None:
        try:
            port = _find_available_port(args.host, 7070)
        except RuntimeError as exc:
            print(f"Error: {exc}", file=sys.stderr)
            sys.exit(1)
        if port != 7070:
            print(f"[lenslet] Port 7070 is in use; using {port} instead.")

    if args.no_write:
        if args.cache_wh:
            print("[lenslet] --no-write disables parquet caching; use --no-cache-wh to silence.")
            args.cache_wh = False

    # Print startup banner
    if is_remote_table:
        mode_label = "Table (hf dataset)" if remote_kind == "hf" else "Table (remote)"
        display_target = remote_uri or raw_target
    elif is_table_file:
        mode_label = "Table (parquet)"
        display_target = str(target)
    else:
        has_parquet = (target / "items.parquet").is_file()
        mode_label = "Table (items.parquet)" if has_parquet else "In-memory (no files written)"
        display_target = str(target)

    banner_lines = [
        "┌─────────────────────────────────────────────────┐",
        "│                   🔍 Lenslet                    │",
        "│         Lightweight Image Gallery Server        │",
        "├─────────────────────────────────────────────────┤",
        f"│  Target:    {display_target[:35]:<35} │",
        f"│  Server:    http://{args.host}:{port:<24} │",
    ]
    if args.share:
        banner_lines.append("│  Share:     starting...                         │")
    no_write_label = (
        "temp cache" if args.no_write else ("remote" if is_remote_table else "off")
    )
    banner_lines.extend(
        [
            f"│  Mode:      {mode_label:<35} │",
        f"│  No-write:  {no_write_label:<35} │",
            "└─────────────────────────────────────────────────┘",
            "",
        ]
    )
    print("\n".join(banner_lines))

    # Guard against multi-worker mode (in-memory sync is single-process only)
    for env_name in ("UVICORN_WORKERS", "WEB_CONCURRENCY"):
        raw = os.getenv(env_name)
        if not raw:
            continue
        try:
            workers = int(raw)
        except ValueError:
            continue
        if workers > 1:
            print(
                f"[lenslet] Warning: {env_name}={workers} detected. "
                "Collaboration sync requires a single worker to avoid divergence."
            )
            break

    dataset_workspace = None
    preindex_result = None
    preindex_signature = None
    if (not is_remote_table) and (not is_table_file):
        from .workspace import Workspace

        if args.no_write:
            dataset_workspace = Workspace.for_temp_dataset(str(target))
        else:
            dataset_workspace = Workspace.for_dataset(str(target), can_write=True)
        if args.share:
            from .preindex import ensure_local_preindex

            try:
                preindex_result = ensure_local_preindex(target, dataset_workspace)
            except Exception as exc:
                print(f"Error: preindex failed: {exc}", file=sys.stderr)
                sys.exit(1)

            if preindex_result is None:
                print("[lenslet] Preindex skipped: no images found.")
            else:
                if preindex_result.workspace.root != dataset_workspace.root:
                    print(f"[lenslet] Preindex workspace: {preindex_result.workspace.root}")
                if preindex_result.reused:
                    print(f"[lenslet] Preindex cache hit: {preindex_result.image_count} images.")
                else:
                    print(f"[lenslet] Preindex ready: {preindex_result.image_count} images.")
                preindex_signature = preindex_result.signature
                dataset_workspace = preindex_result.workspace
        if args.no_write and dataset_workspace is not None:
            print(
                f"[lenslet] No-write: using temp workspace {dataset_workspace.root} "
                "(thumb cache cap 200 MB)."
            )

    # Start server
    import uvicorn
    from .indexing_status import CliIndexingReporter
    from .server import create_app, create_app_from_storage, create_app_from_table
    from .workspace import Workspace
    indexing_reporter = CliIndexingReporter()
    if is_remote_table:
        try:
            table = _load_remote_table(remote_uri or raw_target)
        except Exception as exc:
            print(f"Error: failed to load remote table '{remote_uri or raw_target}': {exc}", file=sys.stderr)
            sys.exit(1)
        workspace = Workspace.for_dataset(None, can_write=False)
        app = create_app_from_table(
            table=table,
            base_dir=args.base_dir,
            thumb_size=args.thumb_size,
            thumb_quality=args.thumb_quality,
            source_column=args.source_column,
            skip_indexing=args.skip_indexing,
            og_preview=args.og_preview,
            workspace=workspace,
            thumb_cache=args.thumb_cache,
            embedding_config=embedding_config,
            embedding_cache=args.embedding_cache,
            embedding_cache_dir=args.embedding_cache_dir,
            embedding_preload=args.embedding_preload,
            allow_local=False,
            indexing_listener=indexing_reporter.handle_update,
        )
    elif is_table_file:
        base_dir = args.base_dir or str(target.parent)
        storage = _prepare_table_cache(
            parquet_path=target,
            base_dir=base_dir,
            source_column=args.source_column,
            cache_wh=args.cache_wh,
            skip_indexing=args.skip_indexing,
            embedding_config=embedding_config,
        )
        if args.no_write:
            workspace = Workspace.for_temp_dataset(str(target))
        else:
            workspace = Workspace.for_parquet(target, can_write=True)
        app = create_app_from_storage(
            storage,
            show_source=True,
            workspace=workspace,
            thumb_cache=args.thumb_cache,
            og_preview=args.og_preview,
            embedding_parquet_path=str(target),
            embedding_config=embedding_config,
            embedding_cache=args.embedding_cache,
            embedding_cache_dir=args.embedding_cache_dir,
            embedding_preload=args.embedding_preload,
            indexing_listener=indexing_reporter.handle_update,
        )
    else:
        items_path = target / "items.parquet"
        if items_path.is_file() and args.cache_wh:
            _prepare_table_cache(
                parquet_path=items_path,
                base_dir=str(target),
                source_column=args.source_column,
                cache_wh=args.cache_wh,
                skip_indexing=args.skip_indexing,
                quiet=True,
                embedding_config=embedding_config,
            )
        app = create_app(
            root_path=str(target),
            thumb_size=args.thumb_size,
            thumb_quality=args.thumb_quality,
            no_write=args.no_write,
            source_column=args.source_column,
            skip_indexing=args.skip_indexing,
            thumb_cache=args.thumb_cache,
            og_preview=args.og_preview,
            embedding_config=embedding_config,
            embedding_cache=args.embedding_cache,
            embedding_cache_dir=args.embedding_cache_dir,
            embedding_preload=args.embedding_preload,
            indexing_listener=indexing_reporter.handle_update,
            workspace=dataset_workspace,
            preindex_signature=preindex_signature,
        )

    share_tunnel = None
    try:
        if args.share:
            try:
                from .server_browse import warm_recursive_cache
                storage = getattr(app.state, "storage", None)
                browse_cache = getattr(app.state, "recursive_browse_cache", None)
                if storage is not None:
                    print("[lenslet] Warming recursive browse cache...")
                    warmed = warm_recursive_cache(
                        storage,
                        "/",
                        browse_cache,
                        hotpath_metrics=getattr(app.state, "hotpath_metrics", None),
                    )
                    if warmed:
                        print(f"[lenslet] Recursive browse cache ready: {warmed} items.")
            except Exception as exc:
                print(f"[lenslet] Warning: failed to warm recursive browse cache: {exc}")
            try:
                share_tunnel = _start_share_tunnel(port, args.host, args.verbose)
            except Exception as exc:
                print(f"[lenslet] Failed to start share tunnel: {exc}", file=sys.stderr)
        uvicorn.run(
            app,
            host=args.host,
            port=port,
            reload=args.reload,
            log_level="info" if args.verbose else "warning",
        )
    finally:
        if share_tunnel is not None:
            share_tunnel.stop()


def _prepare_table_cache(
    parquet_path: Path,
    base_dir: str,
    source_column: str | None,
    cache_wh: bool,
    skip_indexing: bool,
    quiet: bool = False,
    embedding_config=None,
) -> TableStorage:
    from .storage.table import TableStorage, load_parquet_table
    columns = None
    if embedding_config is not None:
        try:
            from .embeddings.detect import columns_without_embeddings, detect_embeddings
            from .storage.table import load_parquet_schema
            schema = load_parquet_schema(str(parquet_path))
            detection = detect_embeddings(schema, embedding_config)
            columns = columns_without_embeddings(schema, detection)
        except Exception as exc:
            if not quiet:
                print(f"[lenslet] Warning: failed to detect embedding columns: {exc}")
    table = load_parquet_table(str(parquet_path), columns=columns)
    width_name = _find_column_name(table, "width")
    height_name = _find_column_name(table, "height")
    missing = _count_missing_dims(table, width_name, height_name)

    effective_skip = skip_indexing
    if cache_wh and skip_indexing and missing > 0:
        effective_skip = False
        if not quiet:
            print("[lenslet] cache-wh enabled; missing width/height -> indexing to populate cache.")

    storage = TableStorage(
        table=table,
        root=base_dir,
        source_column=source_column,
        skip_indexing=effective_skip,
    )

    if cache_wh:
        updated = _maybe_write_dimensions(
            parquet_path=parquet_path,
            table=table,
            width_name=width_name,
            height_name=height_name,
            row_dims=storage.row_dimensions(),
        )
        if updated and not quiet:
            print(f"[lenslet] Cached width/height into {parquet_path}")

    return storage


def _find_column_name(table, target: str) -> str | None:
    target_lower = target.lower()
    for name in table.schema.names:
        if name.lower() == target_lower:
            return name
    return None


def _count_missing_dims(table, width_name: str | None, height_name: str | None) -> int:
    if width_name is None or height_name is None:
        return table.num_rows
    width = table[width_name].to_pylist()
    height = table[height_name].to_pylist()
    missing = 0
    for w, h in zip(width, height):
        if not _valid_dim(w) or not _valid_dim(h):
            missing += 1
    return missing


def _valid_dim(value) -> bool:
    try:
        return value is not None and int(value) > 0
    except Exception:
        return False


def _maybe_write_dimensions(
    parquet_path: Path,
    table,
    width_name: str | None,
    height_name: str | None,
    row_dims: list[tuple[int, int] | None],
) -> bool:
    import pyarrow as pa
    import pyarrow.parquet as pq

    widths = []
    heights = []
    existing_width = table[width_name].to_pylist() if width_name else [None] * table.num_rows
    existing_height = table[height_name].to_pylist() if height_name else [None] * table.num_rows

    changed = False
    for idx in range(table.num_rows):
        w_existing = existing_width[idx] if idx < len(existing_width) else None
        h_existing = existing_height[idx] if idx < len(existing_height) else None
        w_final = w_existing
        h_final = h_existing
        dims = row_dims[idx] if idx < len(row_dims) else None
        if not _valid_dim(w_existing):
            if dims and dims[0] > 0:
                w_final = dims[0]
                changed = True
        if not _valid_dim(h_existing):
            if dims and dims[1] > 0:
                h_final = dims[1]
                changed = True
        widths.append(w_final)
        heights.append(h_final)

    if not changed and width_name and height_name:
        return False

    width_arr = pa.array(widths, type=pa.int64())
    height_arr = pa.array(heights, type=pa.int64())

    table_out = table
    if width_name is None:
        width_name = "width"
        table_out = table_out.append_column(width_name, width_arr)
    else:
        idx = table_out.schema.get_field_index(width_name)
        table_out = table_out.set_column(idx, width_name, width_arr)

    if height_name is None:
        height_name = "height"
        table_out = table_out.append_column(height_name, height_arr)
    else:
        idx = table_out.schema.get_field_index(height_name)
        table_out = table_out.set_column(idx, height_name, height_arr)

    pq.write_table(table_out, str(parquet_path))
    return True


def _detect_source_column(parquet_path: str, base_dir: str | None, sample_size: int = 50) -> str | None:
    try:
        import pyarrow.parquet as pq
    except ImportError as exc:  # pragma: no cover - optional dependency
        raise ImportError(
            "pyarrow is required for Parquet datasets. Install with: pip install pyarrow"
        ) from exc

    pf = pq.ParquetFile(parquet_path)
    if pf.metadata is not None and pf.metadata.num_rows == 0:
        return None

    columns = pf.schema.names
    if not columns:
        return None

    batch = None
    for candidate in pf.iter_batches(batch_size=sample_size):
        batch = candidate
        break
    if batch is None or batch.num_rows == 0:
        return None

    best_score = 0.0
    best_total = 0
    best_name = None
    for col in columns:
        values = batch.column(col).to_pylist()
        total = 0
        matches = 0
        for value in values:
            if value is None:
                continue
            if isinstance(value, float) and math.isnan(value):
                continue
            if isinstance(value, os.PathLike):
                value = os.fspath(value)
            if not isinstance(value, str):
                continue
            value = value.strip()
            if not value:
                continue
            total += 1
            if _is_loadable_value(value, base_dir):
                matches += 1
        if total == 0:
            continue
        score = matches / total
        if score < 0.7:
            continue
        if score > best_score or (score == best_score and total > best_total):
            best_score = score
            best_total = total
            best_name = col

    return best_name


def _is_loadable_value(value: str, base_dir: str | None) -> bool:
    if value.startswith("s3://"):
        return True
    if value.startswith("http://") or value.startswith("https://"):
        return True
    if os.path.isabs(value):
        return os.path.exists(value)
    if base_dir:
        return os.path.exists(os.path.join(base_dir, value))
    return False


if __name__ == "__main__":
    main()



"""Media and thumbnail response helpers for Lenslet."""

from __future__ import annotations

import asyncio
import os
import stat
from typing import Any, Literal

from fastapi import HTTPException, Request, Response
from fastapi.responses import FileResponse

from .thumb_cache import ThumbCache
from .thumbs import ThumbnailScheduler


class _ClientDisconnected(Exception):
    pass


def _thumb_worker_count() -> int:
    cpu = os.cpu_count() or 2
    return max(1, min(4, cpu))


def _get_cached_thumbnail(storage, path: str) -> bytes | None:
    cache = getattr(storage, "_thumbnails", None)
    if not isinstance(cache, dict):
        return None
    if path in cache:
        return cache[path]
    normalizer = getattr(storage, "_normalize_path", None)
    if callable(normalizer):
        try:
            norm = normalizer(path)
        except Exception:
            return None
        return cache.get(norm)
    return None


async def _await_thumbnail(
    request: Request,
    future,
) -> bytes | None:
    wrapped = asyncio.wrap_future(future)
    while True:
        done, _ = await asyncio.wait({wrapped}, timeout=0.05)
        if done:
            try:
                return wrapped.result()
            except asyncio.CancelledError as exc:
                raise _ClientDisconnected() from exc
        if await request.is_disconnected():
            raise _ClientDisconnected()


def _thumb_cache_source(storage, path: str) -> str | None:
    source = None
    getter = getattr(storage, "get_source_path", None)
    if callable(getter):
        try:
            source = getter(path)
        except Exception:
            source = None
    if source is None:
        mapping = getattr(storage, "_source_paths", None)
        if isinstance(mapping, dict):
            source = mapping.get(path)
            if source is None:
                normalizer = getattr(storage, "_normalize_path", None)
                if callable(normalizer):
                    try:
                        norm = normalizer(path)
                        source = mapping.get(norm)
                    except Exception:
                        source = None
    if not source:
        source = path
    if not _is_remote_source(source):
        root = getattr(storage, "root", None)
        if root and not os.path.isabs(source):
            source = os.path.join(root, source)
    return source


def _is_remote_source(source: str) -> bool:
    return source.startswith("s3://") or source.startswith("http://") or source.startswith("https://")


def _thumb_cache_key(storage, path: str) -> str | None:
    source = _thumb_cache_source(storage, path)
    if not source:
        return None
    size = getattr(storage, "thumb_size", "")
    quality = getattr(storage, "thumb_quality", "")
    parts = [source, str(size), str(quality)]
    if not _is_remote_source(source):
        try:
            etag = storage.etag(path)
        except Exception:
            etag = None
        if etag:
            parts.append(str(etag))
    return "|".join(parts)


def _existing_local_file(source: str) -> tuple[str, os.stat_result] | None:
    source_path = os.path.abspath(os.path.expanduser(source))
    try:
        stat_result = os.stat(source_path)
    except OSError:
        return None
    if not stat.S_ISREG(stat_result.st_mode):
        return None
    return source_path, stat_result


def _resolve_local_file_path(storage, path: str) -> tuple[str, os.stat_result] | None:
    source = _thumb_cache_source(storage, path)
    if not source or _is_remote_source(source):
        return None

    resolver = getattr(storage, "_resolve_local_source", None)
    if callable(resolver):
        try:
            source = resolver(source)
        except Exception:
            return None

    local = getattr(storage, "local", None)
    local_resolver = getattr(local, "resolve_path", None)
    if callable(local_resolver):
        try:
            source = local_resolver(path)
        except Exception:
            pass

    return _existing_local_file(source)


async def _thumb_response_async(
    storage,
    path: str,
    request: Request,
    queue: ThumbnailScheduler,
    thumb_cache: ThumbCache | None = None,
    hotpath_metrics: Any | None = None,
) -> Response:
    cached = _get_cached_thumbnail(storage, path)
    if cached is not None:
        return Response(content=cached, media_type="image/webp")

    cache_key = None
    if thumb_cache is not None:
        cache_key = _thumb_cache_key(storage, path)
        if cache_key:
            cached_disk = thumb_cache.get(cache_key)
            if cached_disk is not None:
                return Response(content=cached_disk, media_type="image/webp")

    future = queue.submit(path, lambda: storage.get_thumbnail(path))
    try:
        thumb = await _await_thumbnail(request, future)
    except _ClientDisconnected:
        cancel_state = queue.cancel(path, future)
        if hotpath_metrics is not None:
            hotpath_metrics.increment("thumb_disconnect_cancel_total")
            if cancel_state in ("queued", "inflight"):
                hotpath_metrics.increment(f"thumb_disconnect_cancel_{cancel_state}_total")
        return Response(status_code=204)

    if thumb is None:
        raise HTTPException(500, "failed to generate thumbnail")
    if thumb_cache is not None and cache_key:
        thumb_cache.set(cache_key, thumb)
    return Response(content=thumb, media_type="image/webp")


FilePrefetchContext = Literal["viewer", "compare"]


def _file_prefetch_context(request: Request | None) -> FilePrefetchContext | None:
    if request is None:
        return None
    raw = (request.headers.get("x-lenslet-prefetch") or "").strip().lower()
    if raw == "viewer":
        return "viewer"
    if raw == "compare":
        return "compare"
    return None


def _file_response(
    storage,
    path: str,
    request: Request | None = None,
    hotpath_metrics: Any | None = None,
) -> Response:
    prefetch_context = _file_prefetch_context(request)
    if prefetch_context is not None and hotpath_metrics is not None:
        hotpath_metrics.increment(f"file_prefetch_{prefetch_context}_total")

    media_type = storage._guess_mime(path)
    local_hit = _resolve_local_file_path(storage, path)
    if local_hit is not None:
        local_path, stat_result = local_hit
        if hotpath_metrics is not None:
            hotpath_metrics.increment("file_response_local_stream_total")
        return FileResponse(path=local_path, media_type=media_type, stat_result=stat_result)
    if hotpath_metrics is not None:
        hotpath_metrics.increment("file_response_fallback_bytes_total")
    data = storage.read_bytes(path)
    return Response(content=data, media_type=media_type)



"""Recursive browse cache helpers for deterministic recursive pagination."""

from __future__ import annotations

import gzip
import hashlib
import json
import os
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Iterable

CACHE_SCHEMA_VERSION = 1
DEFAULT_BROWSE_CACHE_CAP_BYTES = 200 * 1024 * 1024
DEFAULT_BROWSE_MEMORY_ENTRY_LIMIT = 6


def _canonical_scope(path: str) -> str:
    value = (path or "").replace("\\", "/").strip()
    if not value:
        return "/"
    value = "/" + value.lstrip("/")
    if value != "/":
        value = value.rstrip("/")
    return value


def _coerce_int(value: Any) -> int:
    try:
        return int(value)
    except (TypeError, ValueError):
        return 0


def _coerce_float(value: Any) -> float:
    try:
        return float(value)
    except (TypeError, ValueError):
        return 0.0


def _normalize_metrics(raw: Any) -> dict[str, float] | None:
    if not isinstance(raw, dict):
        return None
    metrics: dict[str, float] = {}
    for key, value in raw.items():
        try:
            metrics[str(key)] = float(value)
        except (TypeError, ValueError):
            continue
    return metrics or None


def _scopes_overlap(scope: str, changed: str) -> bool:
    if changed == "/" or scope == "/":
        return True
    if scope == changed:
        return True
    if scope.startswith(changed + "/"):
        return True
    return changed.startswith(scope + "/")


@dataclass(frozen=True)
class RecursiveCachedItemSnapshot:
    path: str
    name: str
    mime: str
    width: int
    height: int
    size: int
    mtime: float
    url: str | None = None
    source: str | None = None
    metrics: dict[str, float] | None = None

    @classmethod
    def from_cached_item(cls, cached: Any) -> "RecursiveCachedItemSnapshot":
        return cls(
            path=_canonical_scope(str(getattr(cached, "path", ""))),
            name=str(getattr(cached, "name", "")),
            mime=str(getattr(cached, "mime", "image/jpeg")),
            width=_coerce_int(getattr(cached, "width", 0)),
            height=_coerce_int(getattr(cached, "height", 0)),
            size=_coerce_int(getattr(cached, "size", 0)),
            mtime=_coerce_float(getattr(cached, "mtime", 0)),
            url=_coerce_optional_text(getattr(cached, "url", None)),
            source=_coerce_optional_text(getattr(cached, "source", None)),
            metrics=_normalize_metrics(getattr(cached, "metrics", None)),
        )

    @classmethod
    def from_payload(cls, payload: Any) -> "RecursiveCachedItemSnapshot":
        if not isinstance(payload, dict):
            raise ValueError("invalid cached item payload")
        return cls(
            path=_canonical_scope(str(payload.get("path", ""))),
            name=str(payload.get("name", "")),
            mime=str(payload.get("mime", "image/jpeg")),
            width=_coerce_int(payload.get("width", 0)),
            height=_coerce_int(payload.get("height", 0)),
            size=_coerce_int(payload.get("size", 0)),
            mtime=_coerce_float(payload.get("mtime", 0)),
            url=_coerce_optional_text(payload.get("url")),
            source=_coerce_optional_text(payload.get("source")),
            metrics=_normalize_metrics(payload.get("metrics")),
        )

    def to_payload(self) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "path": self.path,
            "name": self.name,
            "mime": self.mime,
            "width": self.width,
            "height": self.height,
            "size": self.size,
            "mtime": self.mtime,
        }
        if self.url:
            payload["url"] = self.url
        if self.source:
            payload["source"] = self.source
        if self.metrics is not None:
            payload["metrics"] = self.metrics
        return payload


def _coerce_optional_text(value: Any) -> str | None:
    if value is None:
        return None
    text = str(value).strip()
    return text or None


@dataclass(frozen=True)
class RecursiveSnapshotWindow:
    scope_path: str
    sort_mode: str
    generation: str
    items: tuple[RecursiveCachedItemSnapshot, ...]

    @property
    def total_items(self) -> int:
        return len(self.items)


class RecursiveBrowseCache:
    """Hybrid in-memory/on-disk cache for recursive browse windows."""

    def __init__(
        self,
        *,
        cache_dir: Path | None = None,
        max_disk_bytes: int = DEFAULT_BROWSE_CACHE_CAP_BYTES,
        max_memory_entries: int = DEFAULT_BROWSE_MEMORY_ENTRY_LIMIT,
    ) -> None:
        self._lock = threading.Lock()
        self._memory: dict[tuple[str, str, str], RecursiveSnapshotWindow] = {}
        self._memory_access: dict[tuple[str, str, str], float] = {}
        self._max_memory_entries = max(1, int(max_memory_entries))
        self._max_disk_bytes = max(0, int(max_disk_bytes))
        self._cache_dir = Path(cache_dir) if cache_dir is not None else None
        self._persistence_enabled = self._enable_persistence()
        self._warm_executor = ThreadPoolExecutor(
            max_workers=1,
            thread_name_prefix="lenslet-recursive-cache-warm",
        )
        self._warm_jobs: dict[tuple[str, str, str], threading.Event] = {}
        self._persist_jobs: set[tuple[str, str, str]] = set()
        if self._persistence_enabled:
            self._evict_disk_to_cap()

    @property
    def cache_dir(self) -> Path | None:
        return self._cache_dir

    @property
    def persistence_enabled(self) -> bool:
        return self._persistence_enabled

    @property
    def max_disk_bytes(self) -> int:
        return self._max_disk_bytes

    def pending_warm_count(self) -> int:
        with self._lock:
            return len(self._warm_jobs)

    def load(
        self,
        scope_path: str,
        sort_mode: str,
        generation: str,
    ) -> tuple[RecursiveSnapshotWindow, str] | None:
        key = self._cache_key(scope_path, sort_mode, generation)
        with self._lock:
            cached = self._memory.get(key)
            if cached is not None:
                self._memory_access[key] = time.monotonic()
                return cached, "memory"

        if not self._persistence_enabled:
            return None

        loaded = self._load_disk_window(scope_path, sort_mode, generation)
        if loaded is None:
            return None

        with self._lock:
            self._insert_memory_locked(key, loaded)
        return loaded, "disk"

    def save(
        self,
        scope_path: str,
        sort_mode: str,
        generation: str,
        items: Iterable[RecursiveCachedItemSnapshot],
        *,
        defer_persist: bool = False,
    ) -> tuple[RecursiveSnapshotWindow, bool]:
        scope = _canonical_scope(scope_path)
        snapshots = tuple(items)
        window = RecursiveSnapshotWindow(
            scope_path=scope,
            sort_mode=sort_mode,
            generation=generation,
            items=snapshots,
        )
        key = self._cache_key(scope_path, sort_mode, generation)
        with self._lock:
            self._insert_memory_locked(key, window)
        wrote_disk = False
        if defer_persist and self._persistence_enabled:
            wrote_disk = self._schedule_persist_write(key, window)
        else:
            wrote_disk = self._save_disk_window(window)
        return window, wrote_disk

    def invalidate_path(self, path: str | None = None) -> None:
        if path is None:
            with self._lock:
                self._cancel_pending_warms_locked(None)
                self._memory.clear()
                self._memory_access.clear()
            self._clear_disk()
            return

        canonical = _canonical_scope(path)
        with self._lock:
            for key in list(self._memory.keys()):
                scope, _sort_mode, _generation = key
                if _scopes_overlap(scope, canonical):
                    self._memory.pop(key, None)
                    self._memory_access.pop(key, None)
            self._cancel_pending_warms_locked(canonical)

        # Path-scoped disk invalidation would require parsing full entry payloads.
        # For now we clear persisted entries to avoid stale ancestor/descendant reuse.
        self._clear_disk()

    def clear(self) -> None:
        self.invalidate_path(None)

    def schedule_warm(
        self,
        scope_path: str,
        sort_mode: str,
        generation: str,
        producer: Callable[[threading.Event], Iterable[RecursiveCachedItemSnapshot] | None],
        *,
        delay_seconds: float = 0.0,
        on_saved: Callable[[bool], None] | None = None,
    ) -> bool:
        key = self._cache_key(scope_path, sort_mode, generation)
        with self._lock:
            if key in self._memory or key in self._warm_jobs:
                return False
            cancel_event = threading.Event()
            self._warm_jobs[key] = cancel_event

        def _run() -> None:
            wrote_persisted = False
            try:
                if delay_seconds > 0 and cancel_event.wait(timeout=delay_seconds):
                    return
                snapshots = producer(cancel_event)
                if snapshots is None or cancel_event.is_set():
                    return
                _window, wrote_persisted = self.save(scope_path, sort_mode, generation, snapshots)
            except Exception:
                return
            finally:
                if on_saved is not None:
                    try:
                        on_saved(wrote_persisted)
                    except Exception:
                        pass
                with self._lock:
                    self._warm_jobs.pop(key, None)

        try:
            self._warm_executor.submit(_run)
            return True
        except Exception:
            with self._lock:
                self._warm_jobs.pop(key, None)
            return False

    def disk_usage_bytes(self) -> int:
        if not self._persistence_enabled:
            return 0
        total = 0
        for path in self._iter_cache_files():
            try:
                total += path.stat().st_size
            except OSError:
                continue
        return total

    def _cache_key(self, scope_path: str, sort_mode: str, generation: str) -> tuple[str, str, str]:
        return (_canonical_scope(scope_path), sort_mode, generation)

    def _insert_memory_locked(
        self,
        key: tuple[str, str, str],
        window: RecursiveSnapshotWindow,
    ) -> None:
        self._memory[key] = window
        self._memory_access[key] = time.monotonic()
        while len(self._memory) > self._max_memory_entries:
            oldest_key = min(self._memory_access, key=self._memory_access.get)
            self._memory.pop(oldest_key, None)
            self._memory_access.pop(oldest_key, None)

    def _cancel_pending_warms_locked(self, path: str | None) -> None:
        if path is None:
            keys = list(self._warm_jobs.keys())
        else:
            canonical = _canonical_scope(path)
            keys = [key for key in self._warm_jobs.keys() if _scopes_overlap(key[0], canonical)]
        for key in keys:
            event = self._warm_jobs.pop(key, None)
            if event is not None:
                event.set()

    def _schedule_persist_write(
        self,
        key: tuple[str, str, str],
        window: RecursiveSnapshotWindow,
    ) -> bool:
        with self._lock:
            if key in self._persist_jobs:
                return False
            self._persist_jobs.add(key)

        def _run() -> None:
            try:
                self._save_disk_window(window)
            finally:
                with self._lock:
                    self._persist_jobs.discard(key)

        try:
            self._warm_executor.submit(_run)
            return False
        except Exception:
            with self._lock:
                self._persist_jobs.discard(key)
            return False

    def _enable_persistence(self) -> bool:
        if self._cache_dir is None:
            return False
        try:
            self._cache_dir.mkdir(parents=True, exist_ok=True)
            probe = self._cache_dir / ".browse-cache-write-probe"
            probe.write_text("ok", encoding="utf-8")
            probe.unlink(missing_ok=True)
            return True
        except Exception:
            return False

    def _disk_path_for(self, scope_path: str, sort_mode: str, generation: str) -> Path:
        payload = json.dumps(
            {
                "scope": _canonical_scope(scope_path),
                "sort": sort_mode,
                "generation": generation,
                "schema": CACHE_SCHEMA_VERSION,
            },
            separators=(",", ":"),
            sort_keys=True,
            ensure_ascii=True,
        ).encode("utf-8")
        digest = hashlib.sha256(payload).hexdigest()
        assert self._cache_dir is not None
        return self._cache_dir / digest[:2] / f"{digest}.json.gz"

    def _load_disk_window(
        self,
        scope_path: str,
        sort_mode: str,
        generation: str,
    ) -> RecursiveSnapshotWindow | None:
        try:
            path = self._disk_path_for(scope_path, sort_mode, generation)
        except Exception:
            return None
        if not path.is_file():
            return None
        try:
            with gzip.open(path, "rt", encoding="utf-8") as handle:
                payload = json.load(handle)
        except Exception:
            return None

        if not isinstance(payload, dict):
            self._safe_unlink(path)
            return None
        if payload.get("schema_version") != CACHE_SCHEMA_VERSION:
            self._safe_unlink(path)
            return None
        if _canonical_scope(str(payload.get("scope_path", ""))) != _canonical_scope(scope_path):
            self._safe_unlink(path)
            return None
        if str(payload.get("sort_mode", "")) != sort_mode:
            self._safe_unlink(path)
            return None
        if str(payload.get("generation", "")) != generation:
            return None

        raw_items = payload.get("items")
        if not isinstance(raw_items, list):
            self._safe_unlink(path)
            return None

        snapshots: list[RecursiveCachedItemSnapshot] = []
        try:
            for entry in raw_items:
                snapshots.append(RecursiveCachedItemSnapshot.from_payload(entry))
        except Exception:
            self._safe_unlink(path)
            return None

        try:
            os.utime(path, None)
        except OSError:
            pass

        return RecursiveSnapshotWindow(
            scope_path=_canonical_scope(scope_path),
            sort_mode=sort_mode,
            generation=generation,
            items=tuple(snapshots),
        )

    def _save_disk_window(self, window: RecursiveSnapshotWindow) -> bool:
        if not self._persistence_enabled:
            return False
        tmp: Path | None = None
        try:
            path = self._disk_path_for(window.scope_path, window.sort_mode, window.generation)
            path.parent.mkdir(parents=True, exist_ok=True)
            tmp = path.with_name(f"{path.name}.{os.getpid()}.tmp")
            payload = {
                "schema_version": CACHE_SCHEMA_VERSION,
                "scope_path": window.scope_path,
                "sort_mode": window.sort_mode,
                "generation": window.generation,
                "items": [item.to_payload() for item in window.items],
            }
            with gzip.open(tmp, "wt", encoding="utf-8") as handle:
                json.dump(payload, handle, separators=(",", ":"), sort_keys=True)
            tmp.replace(path)
            self._evict_disk_to_cap()
            return path.exists()
        except Exception:
            if tmp is not None:
                try:
                    tmp.unlink(missing_ok=True)
                except Exception:
                    pass
            return False

    def _evict_disk_to_cap(self) -> None:
        if not self._persistence_enabled or self._max_disk_bytes <= 0:
            return

        entries: list[tuple[float, int, Path]] = []
        total = 0
        for path in self._iter_cache_files():
            try:
                stat = path.stat()
            except OSError:
                continue
            total += stat.st_size
            entries.append((stat.st_mtime, stat.st_size, path))

        if total <= self._max_disk_bytes:
            return

        entries.sort(key=lambda entry: entry[0])
        for _mtime, size, path in entries:
            self._safe_unlink(path)
            total -= size
            if total <= self._max_disk_bytes:
                break
        self._cleanup_empty_dirs()

    def _iter_cache_files(self) -> list[Path]:
        if self._cache_dir is None or not self._cache_dir.exists():
            return []
        return [path for path in self._cache_dir.rglob("*.json.gz") if path.is_file()]

    def _clear_disk(self) -> None:
        if not self._persistence_enabled:
            return
        for path in self._iter_cache_files():
            self._safe_unlink(path)
        self._cleanup_empty_dirs()

    def _safe_unlink(self, path: Path) -> None:
        try:
            path.unlink()
        except FileNotFoundError:
            return
        except Exception:
            return

    def _cleanup_empty_dirs(self) -> None:
        if self._cache_dir is None or not self._cache_dir.exists():
            return
        try:
            candidates = sorted(self._cache_dir.rglob("*"), reverse=True)
        except Exception:
            return
        for path in candidates:
            if not path.is_dir():
                continue
            try:
                path.rmdir()
            except OSError:
                continue
